from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QEvent, QMarginsF, QPoint, QPointF, QRectF, Qt
from PySide6.QtGui import QImage, QMouseEvent, QPainter, QPainterPath, QPen, QWheelEvent
from PySide6.QtWidgets import QFrame, QGraphicsItem, QGraphicsScene, QGraphicsView

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from .qgraph import QBaseGraphicsView


def clamp(v, min_, max_):
    return min(max(v, min_), max_)


class QMiniMapViewportBox(QGraphicsItem):
    """
    Widget to indicate target viewport position/size on the minimap.
    """

    PEN_WIDTH: float = 2

    def __init__(self) -> None:
        super().__init__()
        self._scene_rect: QRectF = QRectF()
        self._viewport_rect: QRectF = QRectF()

    def set_scene_rect(self, rect: QRectF) -> None:
        """
        Define the dimensions of the total minimap scene, to render the outer border.
        """
        self.prepareGeometryChange()
        self._scene_rect = rect
        self.update()

    def set_viewport_rect(self, rect: QRectF) -> None:
        """
        Define the offset and dimensions of the displayed viewport indicator in the minimap scene.
        """
        self._viewport_rect = rect
        self.update()

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        """
        Draw the minimap outline and viewport box.
        """
        # Minimap outline
        painter.setPen(QPen(Conf.disasm_view_minimap_outline_color, self.PEN_WIDTH))
        path = QPainterPath()
        path.addRect(self._scene_rect)
        painter.drawPath(path)

        # Viewport box
        painter.setPen(QPen(Conf.disasm_view_minimap_viewport_color, self.PEN_WIDTH))
        path = QPainterPath()
        path.addRect(self._viewport_rect)
        painter.drawPath(path)

    def boundingRect(self):
        half_pen_width = self.PEN_WIDTH / 2
        margins = QMarginsF(half_pen_width, half_pen_width, half_pen_width, half_pen_width)
        return self._scene_rect.marginsAdded(margins)


class QMiniMapTargetSceneViewer(QGraphicsItem):
    """
    Widget to render minimized version of the target view's scene on the minimap.

    For performance, the scene is cached in a QImage and only updated when update_scene_drawing is called.
    """

    def __init__(self, target_view: QGraphicsView) -> None:
        super().__init__()
        self._view: QGraphicsView = target_view
        self._minimap_scene_rect: QRectF = QRectF()
        self._scene_img: QImage = QImage()

    def set_scene_rect(self, rect: QRectF) -> None:
        """
        Define the dimensions of the total minimap scene.
        """
        self.prepareGeometryChange()
        self._minimap_scene_rect = rect
        self.update_scene_drawing()
        self.update()

    def update_scene_drawing(self) -> None:
        """
        Render the target scene in an image to be used for minimap painting.
        """
        scene = self._view.scene()
        if scene is None:
            return

        dpr = self._view.devicePixelRatioF()
        self._scene_img = QImage(
            dpr * self._minimap_scene_rect.width(), dpr * self._minimap_scene_rect.height(), QImage.Format.Format_ARGB32
        )
        self._scene_img.setDevicePixelRatio(dpr)
        self._scene_img.fill(Conf.disasm_view_minimap_background_color)
        self._view.set_extra_render_pass(True)
        painter = QPainter(self._scene_img)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)
        scene.render(painter, target=self._minimap_scene_rect)
        self._view.set_extra_render_pass(False)
        self.update()

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        """
        Paint the minimized scene image.
        """
        painter.drawImage(0, 0, self._scene_img)

    def boundingRect(self):
        return self._minimap_scene_rect


class QMiniMapView(QGraphicsView):
    """
    Renders a minimized version of a QBaseGraphicsView, indicating current viewport and handling mouse events to
    support viewport control.
    """

    def __init__(self, target_view: QBaseGraphicsView, parent=None) -> None:
        super().__init__(parent=parent)
        self._target_view: QBaseGraphicsView = target_view
        self._is_mouse_pressed: bool = False
        self._scale: float = 1.0
        self._minimap_scene: QGraphicsScene = QGraphicsScene()
        self.setScene(self._minimap_scene)
        self._minimap_target_viewport_box: QMiniMapViewportBox = QMiniMapViewportBox()
        self._minimap_scene.addItem(self._minimap_target_viewport_box)
        self._minimap_target_viewport_box.setZValue(1.0)
        self._minimap_target_scene_viewer: QMiniMapTargetSceneViewer = QMiniMapTargetSceneViewer(self._target_view)
        self._minimap_scene.addItem(self._minimap_target_scene_viewer)
        self._target_view.visible_scene_rect_changed.connect(self.on_target_viewport_visible_scene_rect_changed)
        self.setFrameStyle(QFrame.Shape.NoFrame)

    def on_target_viewport_visible_scene_rect_changed(self, visible: QRectF) -> None:
        scene = self._target_view.scene()
        if scene is None:
            return

        x = (visible.x() - scene.sceneRect().x()) * self._scale
        y = (visible.y() - scene.sceneRect().y()) * self._scale
        width = visible.width() * self._scale
        height = visible.height() * self._scale
        mm_scene_w = self.sceneRect().width()
        mm_scene_h = self.sceneRect().height()

        minimap_vp_rect = QRectF()
        minimap_vp_rect.setTopLeft(QPoint(int(clamp(x, 0, mm_scene_w)), int(clamp(y, 0, mm_scene_h))))
        minimap_vp_rect.setBottomRight(
            QPoint(int(clamp(x + width, 0, mm_scene_w)), int(clamp(y + height, 0, mm_scene_h)))
        )

        self._minimap_target_viewport_box.set_viewport_rect(minimap_vp_rect)
        self._minimap_target_viewport_box.update()

    def reload_target_scene(self) -> None:
        """
        Reload scene from target view, scaling the minimap view to properly fit the scene into view while preserving
        scene aspect ratio.
        """
        scene = self._target_view.scene()
        if scene is None:
            return

        # Scale target scene dimensions to fit within widget bounds, preserving scene aspect ratio
        mm_max_w = self.maximumWidth()
        mm_max_h = self.maximumHeight()
        scene_rect = scene.sceneRect()
        scene_w = scene_rect.width()
        scene_h = scene_rect.height()

        if mm_max_w == 0 or mm_max_h == 0 or scene_w == 0 or scene_h == 0:
            return

        minimap_aspect_ratio = mm_max_w / mm_max_h
        scene_aspect_ratio = scene_w / scene_h

        if minimap_aspect_ratio < scene_aspect_ratio:
            self._scale = mm_max_w / scene_w
        else:
            self._scale = mm_max_h / scene_h

        scaled_scene_rect = QRectF(0, 0, int(scene_w * self._scale), int(scene_h * self._scale))
        self.resize(scaled_scene_rect.width(), scaled_scene_rect.height())
        self._minimap_scene.setSceneRect(scaled_scene_rect)
        self.setSceneRect(scaled_scene_rect)
        self._minimap_target_scene_viewer.set_scene_rect(scaled_scene_rect)
        self._minimap_target_viewport_box.set_scene_rect(scaled_scene_rect)

        self._minimap_target_scene_viewer.update_scene_drawing()
        self.on_target_viewport_visible_scene_rect_changed(self._target_view.visible_scene_rect)

    def map_event_pos_to_target_scene_pos(self, pos: QPoint) -> QPointF:
        """
        Map a point `pos` from view (e.g. mouse event) to target scene.
        """
        scene_rect = self._target_view.scene().sceneRect()
        pos = self.mapToScene(pos)
        x, y = pos.x(), pos.y()
        x = x / self._scale + scene_rect.x()
        y = y / self._scale + scene_rect.y()
        return QPointF(x, y)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """
        Handle mouse press, moving the target view port to cursor position in target scene.
        """
        if event.button() == Qt.MouseButton.LeftButton:
            self._is_mouse_pressed = True
            self.setCursor(Qt.CursorShape.ClosedHandCursor)
            self._target_view.centerOn(self.map_event_pos_to_target_scene_pos(event.pos()))
            event.accept()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """
        Handle mouse move, moving the target view port to cursor position in target scene when mouse is pressed.
        """
        if self._is_mouse_pressed:
            self._target_view.centerOn(self.map_event_pos_to_target_scene_pos(event.pos()))
            event.accept()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """
        Handle mouse release, ending viewport drag.
        """
        if event.button() == Qt.MouseButton.LeftButton:
            self.setCursor(Qt.CursorShape.ArrowCursor)
            event.accept()
        self._is_mouse_pressed = False

    def wheelEvent(self, event: QWheelEvent) -> None:
        """
        Forward the wheel event to target view to handle zoom events.
        """
        if event.modifiers() & Qt.KeyboardModifier.ControlModifier == Qt.KeyboardModifier.ControlModifier:
            pos = event.position()
            self._target_view.centerOn(self.map_event_pos_to_target_scene_pos(QPoint(pos.x(), pos.y())))

        self._target_view.wheelEvent(event)

    def changeEvent(self, event: QEvent) -> None:
        """
        Redraw on color scheme update.
        """
        if event.type() == QEvent.Type.PaletteChange:
            self.reload_target_scene()



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication, QDialog, QDialogButtonBox, QLabel, QPlainTextEdit, QVBoxLayout

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class QCommentTextBox(QPlainTextEdit):
    """
    Multiline text box for comment entry.
    """

    def __init__(self, textchanged_callback=None, textconfirmed_callback=None, parent=None) -> None:
        super().__init__(parent)
        if textchanged_callback is not None:
            self.textChanged.connect(textchanged_callback)
        self._textconfirmed_callback = textconfirmed_callback

    @property
    def text(self):
        text = self.toPlainText()
        if text is not None:
            return text.strip()
        return None

    def keyReleaseEvent(self, event):
        if (
            event.key() == Qt.Key_Return
            and QApplication.keyboardModifiers() == Qt.ControlModifier
            and self._textconfirmed_callback is not None
        ):
            self._textconfirmed_callback()
            return True
        return super().keyReleaseEvent(event)


class SetComment(QDialog):
    """
    Dialog for setting comment.
    """

    def __init__(self, workspace: Workspace, comment_addr, parent=None) -> None:
        super().__init__(parent)

        # initialization
        self._workspace = workspace
        self._comment_addr = comment_addr

        self._comment_textbox: QCommentTextBox

        self.setWindowTitle("Set Comment")
        self.main_layout = QVBoxLayout()
        self._init_widgets()
        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        # name label
        comment_lbl = QLabel(self)
        comment_lbl.setText("Comment text")

        # comment textbox
        comment_txtbox = QCommentTextBox(textconfirmed_callback=self._on_ok_clicked, parent=self)
        if self._comment_addr in self._workspace.main_instance.project.kb.comments:
            comment_txtbox.setPlainText(self._workspace.main_instance.project.kb.comments[self._comment_addr])
            comment_txtbox.selectAll()
        self._comment_textbox = comment_txtbox

        comment_layout = QVBoxLayout()
        comment_layout.addWidget(comment_lbl)
        comment_layout.addWidget(comment_txtbox)
        self.main_layout.addLayout(comment_layout)

        # buttons
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_ok_clicked(self) -> None:
        comment_txt = self._comment_textbox.text
        self._workspace.set_comment(self._comment_addr, comment_txt)
        self.close()



"""
Usage information for QtConsole
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.


gui_reference = """\
=====================
The Jupyter QtConsole
=====================

This console is designed to emulate the look, feel and workflow of a terminal
environment. Beyond this basic design, the console also implements
functionality not currently found in most terminal emulators. Some examples of
these console enhancements are inline syntax highlighting, multiline editing,
inline graphics, and others.

This quick reference document contains the basic information you'll need to
know to make the most efficient use of it.  For the various command line
options available at startup, type ``jupyter qtconsole --help`` at the command
line.


Multiline editing
=================

The graphical console is capable of true multiline editing, but it also tries
to behave intuitively like a terminal when possible.  If you are used to
IPython's old terminal behavior, you should find the transition painless. If
you learn to use a few basic keybindings, the console provides even greater
efficiency.

For single expressions or indented blocks, the console behaves almost like the
IPython terminal: single expressions are immediately evaluated, and indented
blocks are evaluated once a single blank line is entered::

    In [1]: print ("Hello Jupyter!")  # Enter was pressed at the end of the line
    Hello Jupyter!

    In [2]: for num in range(10):
       ...:     print(num)
       ...:
    0 1 2 3 4 5 6 7 8 9

If you want to enter more than one expression in a single input block
(something not possible in the terminal), you can use ``Control-Enter`` at the
end of your first line instead of ``Enter``.  At that point the console goes
into 'cell mode' and even if your inputs are not indented, it will continue
accepting lines until either you enter an extra blank line or
you hit ``Shift-Enter`` (the key binding that forces execution).  When a
multiline cell is entered, the console analyzes it and executes its code producing
an ``Out[n]`` prompt only for the last expression in it, while the rest of the
cell is executed as if it was a script.  An example should clarify this::

    In [3]: x=1  # Hit Ctrl-Enter here
       ...: y=2  # from now on, regular Enter is sufficient
       ...: z=3
       ...: x**2  # This does *not* produce an Out[] value
       ...: x+y+z  # Only the last expression does
       ...:
    Out[3]: 6

The behavior where an extra blank line forces execution is only active if you
are actually typing at the keyboard each line, and is meant to make it mimic
the IPython terminal behavior.  If you paste a long chunk of input (for example
a long script copied form an editor or web browser), it can contain arbitrarily
many intermediate blank lines and they won't cause any problems.  As always,
you can then make it execute by appending a blank line *at the end* or hitting
``Shift-Enter`` anywhere within the cell.

With the up arrow key, you can retrieve previous blocks of input that contain
multiple lines.  You can move inside of a multiline cell like you would in any
text editor.  When you want it executed, the simplest thing to do is to hit the
force execution key, ``Shift-Enter`` (though you can also navigate to the end
and append a blank line by using ``Enter`` twice).

If you are editing a multiline cell and accidentally navigate out of it using the
up or down arrow keys, the console clears the cell and replaces it with the
contents of the cell which the up or down arrow key stopped on.  If you wish to
to undo this action,  perhaps because of an accidental keypress, use the Undo
keybinding, ``Control-z``, to restore the original cell.


Key bindings
============

The Jupyter QtConsole supports most of the basic Emacs line-oriented keybindings,
in addition to some of its own.

The keybindings themselves are:

- ``Enter``: insert new line (may cause execution, see above).
- ``Ctrl-Enter``: *force* new line, *never* causes execution.
- ``Shift-Enter``: *force* execution regardless of where cursor is, no newline added.
- ``Up``: step backwards through the history.
- ``Down``: step forwards through the history.
- ``Shift-Up``: search backwards through the history (like ``Control-r`` in bash).
- ``Shift-Down``: search forwards through the history.
- ``Control-c``: copy highlighted text to clipboard (prompts are automatically stripped).
- ``Control-Shift-c``: copy highlighted text to clipboard (prompts are not stripped).
- ``Control-v``: paste text from clipboard.
- ``Control-z``: undo (retrieves lost text if you move out of a cell with the arrows).
- ``Control-Shift-z``: redo.
- ``Control-o``: move to 'other' area, between pager and terminal.
- ``Control-l``: clear terminal.
- ``Control-a``: go to beginning of line.
- ``Control-e``: go to end of line.
- ``Control-u``: kill from cursor to the begining of the line.
- ``Control-k``: kill from cursor to the end of the line.
- ``Control-y``: yank (paste)
- ``Control-p``: previous line (like up arrow)
- ``Control-n``: next line (like down arrow)
- ``Control-f``: forward (like right arrow)
- ``Control-b``: back (like left arrow)
- ``Control-d``: delete next character, or exits if input is empty
- ``Alt-<``: move to the beginning of the input region.
- ``alt->``: move to the end of the input region.
- ``Alt-d``: delete next word.
- ``Alt-Backspace``: delete previous word.
- ``Control-.``: force a kernel restart (a confirmation dialog appears).
- ``Control-+``: increase font size.
- ``Control--``: decrease font size.
- ``Control-Alt-Space``: toggle full screen. (Command-Control-Space on Mac OS X)

The pager
=========

The Jupyter QtConsole will show long blocks of text from many sources using a
built-in pager. You can control where this pager appears with the ``--paging``
command-line flag:

- ``inside`` [default]: the pager is overlaid on top of the main terminal. You
  must quit the pager to get back to the terminal (similar to how a pager such
  as ``less`` or ``more`` pagers behave).

- ``vsplit``: the console is made double height, and the pager appears on the
  bottom area when needed.  You can view its contents while using the terminal.

- ``hsplit``: the console is made double width, and the pager appears on the
  right area when needed.  You can view its contents while using the terminal.

- ``none``: the console displays output without paging.

If you use the vertical or horizontal paging modes, you can navigate between
terminal and pager as follows:

- Tab key: goes from pager to terminal (but not the other way around).
- Control-o: goes from one to another always.
- Mouse: click on either.

In all cases, the ``q`` or ``Escape`` keys quit the pager (when used with the
focus on the pager area).

Running subprocesses
====================

When running a subprocess from the kernel, you can not interact with it as if
it was running in a terminal.  So anything that invokes a pager or expects
you to type input into it will block and hang (you can kill it with ``Control-C``).

The console can use magics provided by the IPython kernel. These magics include
``%less`` to page files (aliased to ``%more``),
``%clear`` to clear the terminal, and ``%man`` on Linux/OSX.  These cover the
most common commands you'd want to call in your subshell and that would cause
problems if invoked via ``!cmd``, but you need to be aware of this limitation.

Display
=======

For example, if using the IPython kernel, there are functions available for
object display:


    In [4]: from IPython.display import display

    In [5]: from IPython.display import display_png, display_svg

Python objects can simply be passed to these functions and the appropriate
representations will be displayed in the console as long as the objects know
how to compute those representations. The easiest way of teaching objects how
to format themselves in various representations is to define special methods
such as: ``_repr_svg_`` and ``_repr_png_``. IPython's display formatters
can also be given custom formatter functions for various types::

    In [6]: ip = get_ipython()

    In [7]: png_formatter = ip.display_formatter.formatters['image/png']

    In [8]: png_formatter.for_type(Foo, foo_to_png)

For further details, see ``IPython.core.formatters``.
"""



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from PySide6.QtWidgets import QLabel

from angrmanagement.logic.threads import gui_thread_schedule
from angrmanagement.plugins.base_plugin import BasePlugin
from angrmanagement.ui.widgets.qinst_annotation import QActiveCount, QPassthroughCount

if TYPE_CHECKING:
    from angrmanagement.ui.views import DisassemblyView, SymexecView
    from angrmanagement.ui.workspace import Workspace


class ExecutionStatisticsViewer(BasePlugin):
    """
    Add a step_callback hook to count how many state on and passthrough of a particular address.
    return the result in build_qblock_annotations callback and show in the disassmbly view by
    fetch_qblock_annotations function.
    """

    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)
        self.passthrough_counts = defaultdict(int)
        self.addr_to_active_states = defaultdict(list)
        self.returning_to_here_states = defaultdict(list)
        self.bb_addrs = None
        self.instance = self.workspace.main_instance

        if self.symexec_view:
            self.current_simgr = self.symexec_view.current_simgr
            self.current_simgr.am_subscribe(self._on_simgr_selected)
        self._init_widgets()

    @property
    def disasm_view(self) -> DisassemblyView:
        return self.workspace.view_manager.first_view_in_category("disassembly")

    @property
    def symexec_view(self) -> SymexecView:
        return self.workspace.view_manager.first_view_in_category("symexec")

    def count_passthrough(self, simgr) -> None:
        """Prior to stepping the active states, increment the passthrough count on the basic block(s) that will be
        executed next."""
        if self.bb_addrs is None:
            self.bb_addrs = {b.addr for b in self.instance.cfg.nodes()}
        for s in simgr.active:
            for i_addr in s.block().instruction_addrs:
                if i_addr in self.bb_addrs:
                    self.passthrough_counts[i_addr] += 1

    def step_callback(self, simgr) -> None:
        """Called after stepping the simgr"""
        self.count_passthrough(simgr)
        self.count_active_states(simgr)
        gui_thread_schedule(self._refresh_gui)

    def count_active_states(self, simgr) -> None:
        """Count the number of states at each instruction in the program. Do some special work to figure out how many
        states are under each call and syscall instruction."""
        self.addr_to_active_states = defaultdict(list)
        self.returning_to_here_states = defaultdict(list)
        for s in simgr.active:
            # Count states at instruction
            self.addr_to_active_states[s.addr].append(s)
            # Count states under calls
            stack_frame = s.callstack
            while stack_frame:
                self.returning_to_here_states[stack_frame.ret_addr].append(s)
                stack_frame = stack_frame.next
            # Count states in syscalls
            if s.history.jumpkind.startswith("Ijk_Sys"):
                self.addr_to_active_states[s.history.jump_source].append(s)

    def _on_simgr_selected(self, *args, src=None, **kwargs) -> None:  # pylint: disable=unused-argument
        """Listener for when a new simgr is selected in the symexec view"""
        if src in ["clicked", "from above"]:
            # The "from above" event is emitted when you create a new simgr via right click menu in disasm view
            self.passthrough_counts = defaultdict(int)
            self.addr_to_active_states = defaultdict(list)
            self.returning_to_here_states = defaultdict(list)
            if not self.current_simgr.am_none:
                self.count_active_states(self.current_simgr)
            self._refresh_gui()

    def update_active_states_label(self) -> None:
        if not self.current_simgr.am_none:
            self.active_states_label.setText(f"Active states: {len(self.current_simgr.active)}")

    def build_qblock_annotations(self, qblock):
        if self.current_simgr.am_none:
            # If there's no simgr selected at this moment, then there arent' any labels to show? Maybe
            return []

        qinsns = qblock.addr_to_insns.values()

        items = []
        for qinsn in qinsns:
            addr = qinsn.addr
            active_states = []
            passthrough_count = 0
            active_states += self.addr_to_active_states[addr]
            passthrough_count += self.passthrough_counts[addr]
            if qinsn.insn.mnemonic.opcode_string == "call":
                ret_addr = qinsn.insn.addr + qinsn.insn.size
                active_states += self.returning_to_here_states[ret_addr]
            if len(active_states) > 0:
                items.append(QActiveCount(qinsn.addr, active_states))
            if passthrough_count > 0:
                items.append(QPassthroughCount(qinsn.addr, passthrough_count))
        return items

    def _refresh_gui(self) -> None:
        self.update_active_states_label()
        self.disasm_view.refresh()

    def _init_widgets(self) -> None:
        statusbar_layout = self.disasm_view._statusbar.layout()
        label = QLabel("Active states: 0")
        label.setStyleSheet("QLabel {color: green}")
        statusbar_layout.insertWidget(1, label)
        self.active_states_label = label



from __future__ import annotations

import os
import re
import urllib.parse

import requests
from PySide6.QtWidgets import QFileDialog

from angrmanagement.config import Conf
from angrmanagement.errors import InvalidURLError, UnexpectedStatusCodeError


def isurl(uri) -> bool:
    try:
        result = urllib.parse.urlparse(uri)
        if result.scheme in ("http", "https"):
            return True
    except ValueError:
        pass

    return False


def download_url(url, parent=None, to_file: bool = True, file_path=None, use_proxies: bool = True):
    if not isurl(url):
        raise TypeError("The given URL %s is not a valid URL.", url)

    r = urllib.parse.urlparse(url)
    basename = os.path.basename(r.path)
    if use_proxies and (Conf.http_proxy or Conf.https_proxy):
        proxies = {
            "http": Conf.http_proxy,
            "https": Conf.https_proxy,
        }
    else:
        proxies = None

    try:
        header = requests.head(url, allow_redirects=True, proxies=proxies)
    except requests.exceptions.InvalidURL as ex:
        raise InvalidURLError from ex

    if header.status_code != 200:
        raise UnexpectedStatusCodeError(header.status_code)

    if "content-disposition" in header.headers:
        # update the base name
        fnames = re.findall("filename=(.+)", header.headers["content-disposition"])
        if fnames:
            basename = fnames[0].strip('"')

    if to_file:
        # save the content to a file and then return the path
        if file_path is None:
            filename, folder = QFileDialog.getSaveFileName(parent, "Download a file to...", basename, "Any file (*);")
            if filename and folder:
                target_path = os.path.join(folder, filename)
            else:
                # terminated
                return None
        else:
            target_path = file_path

        # downloading it
        req = requests.get(url, allow_redirects=True, proxies=proxies)
        if req.status_code != 200:
            raise UnexpectedStatusCodeError(req.status_code)

        with open(target_path, "wb") as f:
            f.write(req.content)
        return target_path

    else:
        # download the content and return as a blob
        # downloading it
        req = requests.get(url, allow_redirects=True, proxies=proxies)
        if req.status_code != 200:
            raise UnexpectedStatusCodeError(req.status_code)

        return req.content



from __future__ import annotations


class AngrManagementError(Exception):
    """Base class for all errors raised by angr management."""


class InvalidURLError(AngrManagementError):
    """InvalidURLError is raised when an invalid URL is provided."""


class UnexpectedStatusCodeError(AngrManagementError):
    """UnexpectedStatusCodeError is raised when an unexpected status code is
    received from an HTTP request.
    """

    def __init__(self, status_code) -> None:
        super().__init__()
        self.status_code = status_code


class ContainerAlreadyRegisteredError(AngrManagementError):
    """ContainerAlreadyRegisteredError is raised when a container is already
    registered in the container registry.
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class VFGGenerationJob(InstanceJob):
    """A job that runs the VFG analysis for a function at a given address."""

    def __init__(self, instance: Instance, addr: int) -> None:
        super().__init__("VFG generation", instance, on_finish=self._finish)
        self._addr = addr

    def run(self, _: JobContext):
        return self.instance.project.analyses.VFG(function_start=self._addr)

    def _finish(self, result: Any) -> None:
        self.instance.vfgs[self._addr] = result

    def __repr__(self) -> str:
        return f"Generating VFG for function at {self._addr:#x}"



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.analyses.disassembly import Value
from PySide6.QtCore import QRectF, Qt
from PySide6.QtGui import QBrush, QCursor, QPainter
from PySide6.QtWidgets import QApplication, QGraphicsSceneMouseEvent, QGraphicsSimpleTextItem

from angrmanagement.utils import get_comment_for_display, get_string_for_display, should_display_string_label

from .qgraph_object import QCachedGraphicsItem
from .qoperand import QOperand

if TYPE_CHECKING:
    import PySide6

    from angrmanagement.data.instance import Instance


class QInstruction(QCachedGraphicsItem):
    GRAPH_ADDR_SPACING = 20
    GRAPH_MNEMONIC_SPACING = 10
    GRAPH_OPERAND_SPACING = 2
    GRAPH_COMMENT_STRING_SPACING = 10

    INTERSPERSE_ARGS = ", "

    LINEAR_INSTRUCTION_OFFSET = 120
    COMMENT_PREFIX = "// "

    def __init__(
        self, instance: Instance, func_addr, disasm_view, disasm, infodock, insn, out_branch, config, parent=None
    ) -> None:
        super().__init__(parent=parent)

        # initialization
        self.instance = instance
        self.func_addr = func_addr
        self.disasm_view = disasm_view
        self.disasm = disasm
        self.infodock = infodock
        self.variable_manager = infodock.variable_manager
        self.insn = insn
        self.out_branch = out_branch
        self._config = config

        # all "widgets"
        self._addr = None
        self._mnemonic = None
        self._addr_item: QGraphicsSimpleTextItem = None
        self._mnemonic_item: QGraphicsSimpleTextItem = None
        self._operands: list[QOperand] = []
        self._commas: list[QGraphicsSimpleTextItem] = []
        self._string = None
        self._string_item: QGraphicsSimpleTextItem | None = None
        self._comment = None
        self._comment_items: list[QGraphicsSimpleTextItem] | None = None  # one comment per line
        self._legend = None
        self._width = 0
        self._height = 0

        self._init_widgets()

    def contextMenuEvent(self, event: PySide6.QtWidgets.QGraphicsSceneContextMenuEvent) -> None:
        pass

    def mouseDoubleClickEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.infodock.unselect_all_instructions()
        self.infodock.toggle_instruction_selection(self.addr, insn_pos=self.scenePos(), unique=True)
        self.disasm_view._insn_addr_on_context_menu = self.addr
        self.disasm_view._insn_menu.insn_addr = self.addr
        self.disasm_view.popup_patch_dialog()
        event.accept()

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        if self.disasm_view.workspace.plugins.handle_click_insn(self, event):
            # stop handling this event if the event has been handled by a plugin
            event.accept()
        elif event.button() == Qt.MouseButton.LeftButton and QApplication.keyboardModifiers() in (
            Qt.KeyboardModifier.NoModifier,
            Qt.KeyboardModifier.ControlModifier,
        ):
            # toggle selection
            self.infodock.toggle_instruction_selection(
                self.addr,
                insn_pos=self.scenePos(),
                unique=QApplication.keyboardModifiers() != Qt.KeyboardModifier.ControlModifier,
            )
            event.accept()
        elif (
            event.button() == Qt.MouseButton.RightButton
            and QApplication.keyboardModifiers() == Qt.KeyboardModifier.NoModifier
        ):
            if self.addr not in self.infodock.selected_insns:
                self.infodock.toggle_instruction_selection(self.addr, insn_pos=self.scenePos(), unique=True)
            self.disasm_view.instruction_context_menu(self.insn, QCursor.pos())
            event.accept()
        else:
            super().mousePressEvent(event)

    @property
    def addr(self):
        return self.insn.addr

    def _calc_backcolor(self):
        # First we'll check for customizations
        color = self.disasm_view.workspace.plugins.color_insn(self.insn.addr, self.selected, self.disasm_view)
        if color is not None:
            return color

        if self.selected:
            return self._config.disasm_view_node_instruction_selected_background_color

        if not self.instance.patches.am_none:
            patches = self.instance.patches.get_all_patches(self.insn.addr, self.insn.size)
            if len(patches):
                return Qt.darkYellow

        return None  # None here means transparent, reusing the block color

    @property
    def selected(self):
        """
        If this instruction is selected or not.

        :return:    True if it is selected, False otherwise.
        :rtype:     bool
        """

        return self.infodock.is_instruction_selected(self.addr)

    def clear_cache(self) -> None:
        super().clear_cache()
        for obj in self._operands:
            obj.clear_cache()

    def refresh(self) -> None:
        self.load_comment()
        self._init_comments_or_string()

        for operand in self._operands:
            operand.refresh()

        self._layout_items_and_update_size()
        self.recalculate_size()

    def get_operand(self, operand_idx):
        if operand_idx < len(self._operands):
            return self._operands[operand_idx]
        return None

    def load_comment(self) -> None:
        if self.instance.kb is None:
            self._comment = None
        else:
            self._comment = get_comment_for_display(self.instance.kb, self.insn.addr)

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        painter.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)

        # background color
        backcolor = self._calc_backcolor()
        if backcolor is not None:
            painter.setBrush(backcolor)
            painter.setPen(backcolor)
            painter.drawRect(0, 0, self.width, self.height)

        # any plugin instruction rendering passes
        self.disasm_view.workspace.plugins.draw_insn(self, painter)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self.load_comment()
        self._operands.clear()

        # address
        self._addr = f"{self.insn.addr:08x}"
        self._addr_item = QGraphicsSimpleTextItem(self)
        self._addr_item.setBrush(QBrush(self._config.disasm_view_node_address_color))
        self._addr_item.setFont(self._config.disasm_font)
        self._addr_item.setText(self._addr)

        # mnemonic
        self._mnemonic = self.insn.mnemonic.render()[0]
        self._mnemonic_item = QGraphicsSimpleTextItem(self)
        self._mnemonic_item.setFont(self._config.disasm_font)
        self._mnemonic_item.setBrush(self._config.disasm_view_node_mnemonic_color)
        self._mnemonic_item.setText(self._mnemonic)

        # operands
        for i, operand in enumerate(self.insn.operands):
            is_branch_target = self.insn.type in ("branch", "call") and i == self.insn.branch_target_operand
            is_indirect_branch = self.insn.branch_type == "indirect"
            branch_targets = None
            if is_branch_target:
                if self.out_branch is not None:
                    branch_targets = self.out_branch.targets
                else:
                    # it does not create multiple branches. e.g., a call instruction
                    if len(operand.children) == 1 and type(operand.children[0]) is Value:
                        branch_targets = (operand.children[0].val,)
            qoperand = QOperand(
                self.instance,
                self.func_addr,
                self.disasm_view,
                self.disasm,
                self.infodock,
                self.insn,
                operand,
                i,
                is_branch_target,
                is_indirect_branch,
                branch_targets,
                self._config,
                parent=self,
            )
            self._operands.append(qoperand)

        # all commas
        for _ in range(len(self._operands) - 1):
            comma = QGraphicsSimpleTextItem(self.INTERSPERSE_ARGS, self)
            comma.setFont(self._config.disasm_font)
            comma.setBrush(self._config.disasm_view_node_mnemonic_color)
            self._commas.append(comma)

        if should_display_string_label(self.instance.cfg, self.insn.addr, self.instance.project):
            # yes we should display a string label
            self._string = get_string_for_display(self.instance.cfg, self.insn.addr, self.instance.project)
            if self._string is None:
                self._string = "<Unknown>"

        self._init_comments_or_string()

        self._layout_items_and_update_size()

    def _init_comments_or_string(self) -> None:
        # remove existing comments or strings
        if self._comment_items:
            for comm in self._comment_items:
                comm: QGraphicsSimpleTextItem
                comm.setParentItem(None)
            self._comment_items = None
        elif self._string_item is not None:
            self._string_item.setParentItem(None)
            self._string_item = None

        # comment or string - comments have precedence
        if self._comment:
            self._string_item = None
            lines = self._comment.split("\n")
            self._comment_items = []
            for line in lines:
                comment = QGraphicsSimpleTextItem(self.COMMENT_PREFIX + line, self)
                comment.setFont(self._config.disasm_font)
                comment.setBrush(self._config.disasm_view_comment_color)
                self._comment_items.append(comment)
        elif self._string is not None:
            self._comment_items = None
            self._string_item = QGraphicsSimpleTextItem(self._string, self)
            self._string_item.setFont(self._config.disasm_font)
            self._string_item.setBrush(self._config.disasm_view_string_color)

    def _layout_items_and_update_size(self) -> None:
        x, y = 0, 0

        # address
        if self.disasm_view.show_address:
            self._addr_item.setVisible(True)
            self._addr_item.setPos(x, y)
            x += self._addr_item.boundingRect().width() + self.GRAPH_ADDR_SPACING
        else:
            self._addr_item.setVisible(False)

        # mnemonic
        self._mnemonic_item.setPos(x, y)
        x += self._mnemonic_item.boundingRect().width() + self.GRAPH_MNEMONIC_SPACING

        # operands and commas
        for operand, comma in zip(self._operands, self._commas, strict=False):
            operand.setPos(x, y)
            x += operand.boundingRect().width()
            comma.setPos(x, y)
            x += comma.boundingRect().width()

        # the last operand
        if self._operands:
            self._operands[-1].setPos(x, y)
            x += self._operands[-1].boundingRect().width()

        # comments and strings
        if self._comment_items:
            x += self.GRAPH_COMMENT_STRING_SPACING
            max_comment_width = 0
            for comment in self._comment_items:
                comment.setPos(x, y)
                max_comment_width = max(comment.boundingRect().width(), max_comment_width)
                y += comment.boundingRect().height()
            x += max_comment_width
        elif self._string_item is not None:
            x += self.GRAPH_COMMENT_STRING_SPACING
            self._string_item.setPos(x, y)
            x += self._string_item.boundingRect().width()
            y += self._string_item.boundingRect().height()
        else:
            y = self._mnemonic_item.boundingRect().height()

        # update size
        self._width = x
        self._height = y
        self.recalculate_size()

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)



from __future__ import annotations

import base64
import binascii
import urllib.parse


class UrlActionBase:
    """
    The base class for URL actions.
    """

    def __init__(self, target_id: str = None) -> None:
        self.target_id = target_id

    def act(self, daemon_conn=None):
        raise NotImplementedError

    @staticmethod
    def from_params(params):
        action = params.get("action", None)

        if action is None:
            raise ValueError("'action' is not specified.")
        if not isinstance(action, list):
            raise TypeError("Unexpected type of 'action'. Expecting a list.")
        if len(action) != 1:
            raise ValueError("We do not support multiple actions within the same URL.")
        action = action[0]

        if action not in _ACT2CLS:
            raise KeyError(f"Unsupported action {action}.")
        return _ACT2CLS[action]._from_params(params)

    @staticmethod
    def _one_param(param, key):
        v = param.get(key, None)
        if v is None:
            return None
        return v[0]

    @staticmethod
    def str2addr(addr_str):
        if addr_str is None:
            return None
        try:
            addr = int(addr_str)
        except ValueError:
            addr = int(addr_str, 16)
        return addr


class UrlActionOpen(UrlActionBase):
    """
    Implements the open action.
    """

    def __init__(self, bin_path, target_id=None, headless: bool = False) -> None:
        super().__init__(target_id=target_id)
        self.bin_path = bin_path
        self.headless = headless

    def act(self, daemon_conn=None) -> None:
        if self.bin_path is not None:
            daemon_conn.root.open(self.bin_path)
        else:
            print("Incomplete URL: expected `path` parameter")

    @classmethod
    def _from_params(cls, params):
        return cls(
            cls._one_param(params, "path"),
            target_id=cls._one_param(params, "target_id"),
            headless=cls._one_param(params, "headless"),
        )


class UrlActionJumpTo(UrlActionBase):
    """
    Implements the jump-to action.
    """

    def __init__(self, addr: int | None = None, symbol=None, target_id=None) -> None:
        super().__init__(target_id=target_id)
        self.addr = addr
        self.symbol = symbol

    def act(self, daemon_conn=None) -> None:
        daemon_conn.root.jumpto(self.addr, self.symbol, self.target_id)

    @classmethod
    def _from_params(cls, params):
        addr = cls.str2addr(cls._one_param(params, "addr"))

        return cls(
            addr=addr,
            symbol=cls._one_param(params, "symbol"),
            target_id=cls._one_param(params, "target_id"),
        )


class UrlActionCommentAt(UrlActionBase):
    """
    Implements the comment-at action.
    """

    def __init__(self, addr: int, comment: str, target_id=None) -> None:
        super().__init__(target_id)
        self.addr = addr
        self.comment = comment

    def act(self, daemon_conn=None) -> None:
        if self.addr is None or self.comment is None:
            return
        daemon_conn.root.commentat(self.addr, self.comment, self.target_id)

    @classmethod
    def _from_params(cls, params):
        addr = cls.str2addr(cls._one_param(params, "addr"))
        try:
            comment = base64.b64decode(cls._one_param(params, "comment")).decode("utf-8")
        except (TypeError, binascii.Error):
            comment = None

        return cls(
            addr=addr,
            comment=comment,
            target_id=cls._one_param(params, "target_id"),
        )


class UrlActionBinaryAware(UrlActionBase):
    """
    The base class of all binary-aware URL actions.
    """

    def __init__(self, target_id=None, action=None, kwargs=None) -> None:
        super().__init__(target_id)
        self.action = action
        self.kwargs = kwargs

        if not self.target_id:
            raise TypeError("You must provide the target ID.")
        if not self.action:
            raise TypeError("You must provide action.")

    def act(self, daemon_conn=None) -> None:
        daemon_conn.root.custom_binary_aware_action(self.target_id, self.action, self.kwargs)

    @classmethod
    def _from_params(cls, params):
        target_id: str | None = cls._one_param(params, "target_id")
        action = cls._one_param(params, "action")
        kwargs = {}
        for k, v in params.items():
            if k not in {"target_id", "action"}:
                if isinstance(v, list | tuple):
                    kwargs[k] = v[0]
                else:
                    kwargs[k] = v
        return cls(target_id=target_id, action=action, kwargs=kwargs)


_ACT2CLS: dict[str, type[UrlActionBase]] = {
    "open": UrlActionOpen,
    "jumpto": UrlActionJumpTo,
    "commentat": UrlActionCommentAt,
}


def handle_url(url, act: bool = True):
    o = urllib.parse.urlparse(url)
    params = urllib.parse.parse_qs(o.query)

    action = UrlActionBase.from_params(params)
    if act:
        action.act()
    return action


def register_url_action(action: str, action_handler: type[UrlActionBase]) -> None:
    _ACT2CLS[action] = action_handler



from __future__ import annotations

from typing import TYPE_CHECKING

from angrmanagement.plugins import BasePlugin

from .bhinstance import BhInstance
from .run_target_dialog import RunTargetDialog

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class BinharnessPlugin(BasePlugin):
    """BinHarness plugin for angr management."""

    MENU_BUTTONS = [
        "Binharness: Run",
    ]

    def __init__(self, workspace: Workspace):
        super().__init__(workspace)

        workspace.main_instance.register_container("binharness", BhInstance, BhInstance, "Binharness instance")

    # Internal methods

    def _bh(self) -> BhInstance:
        return self.workspace.main_instance.binharness

    def _show_run_target_dialog(self):
        run_target_dialog = RunTargetDialog(self.workspace, self._bh())
        run_target_dialog.exec_()

    # Event handlers

    def handle_project_initialization(self):
        self._bh().load_project(self.workspace.main_instance.project)

    def handle_click_menu(self, idx: int):
        # TODO: Error if no project loaded
        if idx == 0:
            self._show_run_target_dialog()
        else:
            raise NotImplementedError



from __future__ import annotations

from PySide6.QtWidgets import QDialog, QDialogButtonBox, QHBoxLayout, QLabel, QLineEdit, QVBoxLayout


class InputPromptDialog(QDialog):
    """
    A generic dialog to prompt for text input.
    """

    def __init__(self, window_title: str, prompt_text: str, initial_input_text: str = "", parent=None) -> None:
        super().__init__(parent)
        self.prompt_text: str = prompt_text
        self.initial_input_text: str = initial_input_text
        self.input_edt: QLineEdit = None
        self.result = None

        self.setWindowTitle(window_title)
        self.main_layout = QVBoxLayout()
        self._init_widgets()
        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        prompt_lbl = QLabel(self)
        prompt_lbl.setText(self.prompt_text)

        input_edt = QLineEdit(parent=self)
        input_edt.setText(self.initial_input_text)
        input_edt.selectAll()
        self.input_edt = input_edt

        prompt_input_lyt = QHBoxLayout()
        prompt_input_lyt.addWidget(prompt_lbl)
        prompt_input_lyt.addWidget(input_edt)
        self.main_layout.addLayout(prompt_input_lyt)

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        buttons_lyt = QHBoxLayout()
        buttons_lyt.addWidget(buttons)
        self.main_layout.addLayout(buttons_lyt)

    #
    # Event handlers
    #

    def _on_ok_clicked(self) -> None:
        self.result = self.input_edt.text()
        self.close()



from __future__ import annotations

import ctypes
import itertools
import logging
import sys
import time
from queue import Queue
from threading import Thread
from typing import TYPE_CHECKING

from angrmanagement.data.jobs.job import JobState
from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.threads import gui_thread_schedule, gui_thread_schedule_async

if TYPE_CHECKING:
    from collections.abc import Callable

    from angrmanagement.data.jobs.job import Job
    from angrmanagement.ui.workspace import Workspace


log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class JobContext:
    """JobContext is a context object that is passed to each job to allow it to
    report progress and other information back to the JobManager.
    """

    _job_manager: JobManager
    _job: Job

    def __init__(self, job_manager: JobManager, job: Job):
        self._job_manager = job_manager
        self._job = job

    def set_progress(self, percentage: float, text: str | None = None) -> None:
        self._job_manager.callback_job_set_progress(self._job, percentage, text)


class Worker(Thread):
    """Worker is a thread that runs jobs in the background."""

    job_manager: JobManager
    id_: int

    current_job: Job | None

    def __init__(self, job_manager: JobManager, id_: int):
        super().__init__(name=f"angr-management Worker Thread {id_}", daemon=True)
        self.job_manager = job_manager
        self.id_ = id_
        self.current_job = None

    def run(self) -> None:
        while True:
            if self.job_manager.jobs_queue.empty():
                gui_thread_schedule(GlobalInfo.main_window.status_bar.progress_done, args=())

            if (
                any(job.blocking for job in self.job_manager.jobs)
                and GlobalInfo.main_window is not None
                and GlobalInfo.main_window.workspace
            ):
                gui_thread_schedule(GlobalInfo.main_window.status_bar._progress_dialog.hide, args=())

            job = self.job_manager.jobs_queue.get()
            gui_thread_schedule_async(GlobalInfo.main_window.status_bar.progress, args=("Working...", 0.0, True))

            if any(job.blocking for job in self.job_manager.jobs) and GlobalInfo.main_window.isVisible():
                gui_thread_schedule(GlobalInfo.main_window.status_bar._progress_dialog.show, args=())

            # If job cancelled, then skip it
            if job.state == JobState.CANCELLED:
                continue
            try:
                self.current_job = job
                self.job_manager.callback_worker_new_job(self.current_job)
                ctx = JobContext(self.job_manager, job)
                ctx.set_progress(0)
                job.state = JobState.RUNNING
                log.info('Job "%s" started', job.name)
                job.start_at = time.time()
                result = job.run(ctx)
                if job.state != JobState.CANCELLED:
                    self.job_manager.callback_job_complete(self.current_job)
                now = time.time()
                duration = now - job.start_at
                log.info('Job "%s" completed after %.2f seconds', job.name, duration)
            except (Exception, KeyboardInterrupt) as e:  # pylint: disable=broad-except
                sys.last_traceback = e.__traceback__
                self.current_job = None
                if job.state != JobState.CANCELLED:
                    job.state = JobState.FAILED
                log.exception('Exception while running job "%s":', job.name)
                if self.job_manager.job_worker_exception_callback is not None:
                    self.job_manager.job_worker_exception_callback(job, e)
            else:
                job.state = JobState.FINISHED
                self.current_job = None
                self.job_manager.jobs.remove(job)
                gui_thread_schedule_async(job.finish, args=(result,))

    def keyboard_interrupt(self) -> None:
        """Called from the GUI thread when the user presses Ctrl+C or presses a cancel button"""
        # lol. lmao even.
        if self.ident is not None:
            res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
                ctypes.c_long(self.ident), ctypes.py_object(KeyboardInterrupt)
            )
            if res != 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self.ident), 0)
                log.error("Failed to interrupt thread")


class JobManager:
    """JobManager is responsible for managing jobs and running them in a separate thread."""

    workspace: Workspace

    jobs: list[Job]
    jobs_queue: Queue[Job]
    worker_thread: Worker | None

    job_worker_exception_callback: Callable[[Job, BaseException], None] | None
    _job_id_counter: itertools.count

    _gui_last_updated_at: float
    _last_text: str | None

    def __init__(self, workspace: Workspace):
        self.workspace = workspace

        self.jobs = []
        self.jobs_queue = Queue()
        self.worker_thread = None
        self.job_worker_exception_callback = None
        self._job_id_counter = itertools.count()
        self._gui_last_updated_at = 0.0
        self._last_text = None

    def add_job(self, job: Job) -> None:
        self.jobs.append(job)
        if self.workspace.view_manager.first_view_in_category("jobs") is not None:
            self.callback_job_added(job)
        self.jobs_queue.put(job)

        if self.worker_thread is None or not self.worker_thread.is_alive():
            self._start_worker()

    def cancel_job(self, job: Job) -> bool:
        """Cancel a job. Returns True if the job was cancelled, False if it was
        not found or already completed.
        """
        if job.state not in (JobState.PENDING, JobState.RUNNING):
            return False
        if job in self.jobs:
            self.jobs.remove(job)
            job.state = JobState.CANCELLED
            if self.worker_thread is not None and self.worker_thread.current_job == job:
                self.worker_thread.keyboard_interrupt()

            return True
        return False

    def interrupt_current_job(self) -> None:
        """Notify the current running job that the user requested an interrupt. The job may ignore it."""
        # Due to thread scheduling, current_job reference *must* first be saved on the stack. Accessing self.current_job
        # multiple times will lead to a race condition.
        if self.worker_thread is not None:
            current_job = self.worker_thread.current_job
            if current_job:
                self.worker_thread.keyboard_interrupt()

    def join_all_jobs(self, wait_period: float = 2.0) -> None:
        """
        Wait until self.jobs is empty for at least `wait_period` seconds.

        This is because one job may add another job upon completion. We cannot simply wait until self.jobs becomes
        empty.
        """

        last_has_job = time.time()
        while time.time() - last_has_job <= wait_period:
            while self.jobs:
                last_has_job = time.time()
                time.sleep(0.05)

    def _start_worker(self) -> None:
        self.worker_thread = Worker(self, next(self._job_id_counter))
        self.worker_thread.start()

    # Job callbacks

    def callback_job_set_progress(self, job: Job, percentage: float, text: str | None) -> None:
        delta = percentage - job.progress_percentage

        if (delta > 0.02 or self._last_text != text) and time.time() - self._gui_last_updated_at >= 0.1:
            self._gui_last_updated_at = time.time()
            job.progress_percentage = percentage
            status_text = f"{job.name}: {text}" if text else job.name
            gui_thread_schedule_async(GlobalInfo.main_window.status_bar.progress, args=(status_text, percentage))

            # Dynamically update jobs view progress with instance
            if self.workspace.view_manager.first_view_in_category("jobs") is not None:
                self.callback_worker_progress(job)

    def callback_job_added(self, job: Job) -> None:
        """
        This callback adds jobs dynamically to the jobsView
        upon addition of a new job
        """
        if self.workspace.view_manager.first_view_in_category("jobs") is not None:
            jobs_view = self.workspace.view_manager.first_view_in_category("jobs")
            gui_thread_schedule_async(jobs_view.qjobs.add_new_job, args=[job])

    def callback_worker_progress(self, job: Job) -> None:
        """
        This callback modifies the jobsView table to
        change the progress of a job visually
        """
        if self.workspace.view_manager.first_view_in_category("jobs") is not None:
            jobs_view = self.workspace.view_manager.first_view_in_category("jobs")
            gui_thread_schedule_async(jobs_view.qjobs.change_job_progress, args=[job])

    def callback_worker_new_job(self, job: Job) -> None:
        """
        This callback changes the jobsView table to have the table modified
        with modifying the job status as running
        """
        if self.workspace.view_manager.first_view_in_category("jobs") is not None:
            jobs_view = self.workspace.view_manager.first_view_in_category("jobs")
            gui_thread_schedule_async(jobs_view.qjobs.change_job_running, args=(job,))

    def callback_job_complete(self, job: Job):
        """
        This callback changes the jobsView table to have the table modified
        with the job complete
        """
        if self.workspace.view_manager.first_view_in_category("jobs") is not None:
            jobs_view = self.workspace.view_manager.first_view_in_category("jobs")
            gui_thread_schedule_async(jobs_view.qjobs.change_job_finish, args=[job])

    # Private methods



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QHeaderView,
    QMenu,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance


class EnvTable(QTableWidget):
    """
    Environment Config Table
    """

    def __init__(self, items, parent) -> None:
        super().__init__(parent)

        header_labels = ["Name", "Value"]

        self.setColumnCount(len(header_labels))
        self.setHorizontalHeaderLabels(header_labels)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectItems)
        header = self.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)

        self.setRowCount(len(items))
        for idx, item in enumerate(items):
            for i, it in enumerate(item):
                self.setItem(idx, i, QTableWidgetItem(it))

    def contextMenuEvent(self, event) -> None:
        menu = QMenu("", self)

        menu.addAction("Add a Row", self._action_new_row)
        menu.addAction("Delete this Row", self._action_delete)

        menu.exec_(event.globalPos())

    def _action_new_row(self) -> None:
        row = self.rowCount()
        self.insertRow(row)
        self.setItem(row, 0, QTableWidgetItem("change me"))
        self.setItem(row, 1, QTableWidgetItem(""))

    def _action_delete(self) -> None:
        self.removeRow(self.currentRow())

    def get_result(self):
        ret = []
        for i in range(self.rowCount()):
            ret.append([self.item(i, 0).text(), self.item(i, 1).text()])
        return ret


class EnvConfig(QDialog):
    """
    Environment Config Dialog for new state
    """

    def __init__(self, env_config=None, instance: Instance | None = None, parent=None) -> None:
        super().__init__(parent)

        self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint)
        self._instance = instance
        self._parent = parent
        self.env_config = env_config or []
        self._init_widgets()

    def _init_widgets(self) -> None:
        layout = QVBoxLayout()
        self._table = EnvTable(self.env_config, self)
        layout.addWidget(self._table, 0)
        self.setLayout(layout)

    def closeEvent(self, event) -> None:  # pylint: disable=unused-argument
        self.env_config = self._table.get_result()
        self.close()



from __future__ import annotations

from .info_dock import InfoDock, OperandHighlightMode
from .jump_history import JumpHistory

__all__ = [
    "InfoDock",
    "JumpHistory",
    "OperandHighlightMode",
]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import networkx
from PySide6.QtCore import QSize
from PySide6.QtWidgets import QFrame, QHBoxLayout

from .qstate_block import QStateBlock
from .qsymexec_graph import QSymExecGraph

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace

log = logging.getLogger(__name__)


class QPathTree(QFrame):
    def __init__(self, simgr, state, symexec_view, workspace: Workspace, parent=None) -> None:
        super().__init__(parent=parent)

        self.symexec_view = symexec_view
        self.workspace = workspace
        self.simgr = simgr
        self.state = state

        # widgets
        self._graph = None

        self._init_widgets()

        self.simgr.am_subscribe(self._watch_simgr)

    #
    # Public methods
    #

    def reload(self) -> None:
        if self.simgr.am_none:
            return

        states = [state for (stash, states) in self.simgr.stashes.items() if stash != "pruned" for state in states]
        hierarchy = self.simgr._hierarchy

        graph = self._generate_graph([state.history for state in states], hierarchy, self.symexec_view)

        self._graph.graph = graph

    #
    # Initialization
    #

    def _init_widgets(self) -> None:
        graph = QSymExecGraph(self.state, self.workspace, self.symexec_view, parent=self)

        self._graph = graph

        layout = QHBoxLayout()
        layout.addWidget(graph)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    #
    # Overriden methods
    #

    def sizeHint(self):
        return QSize(500, 500)

    #
    # Private methods
    #

    @staticmethod
    def _all_paths(paths, hierarchy):
        work = set(paths)
        seen = set()
        while len(work) > 0:
            path = work.pop()
            if (
                not hierarchy.history_contains(path.history)
                or len(hierarchy.history_successors(path.history)) == 0
                and path.path_id not in seen
            ):
                yield path
                seen.add(path.path_id)
            # get parents
            if hierarchy.history_contains(path.history):
                parents = hierarchy.history_predecessors(path.history)
                for parent_history in parents:
                    # assume _path_mapping always has the path
                    parent_path = hierarchy._path_mapping[parent_history]
                    work.add(parent_path)
                    if len(hierarchy.history_successors(parent_history)) > 1 and parent_path.path_id not in seen:
                        yield parent_path
                        seen.add(parent_path.path_id)

    @staticmethod
    def _all_edges_gen(state_histories, hierarchy):
        # TODO: reduce duplication with above function
        work = set(state_histories)
        while len(work) > 0:
            working_history = bot_history = work.pop()
            while hierarchy.history_contains(working_history):
                parent_histories = hierarchy.history_predecessors(working_history)
                if not parent_histories:
                    break

                parent_history = parent_histories[0]

                try:
                    successors = hierarchy.history_successors(parent_history)
                    if len(successors) > 1:
                        yield (parent_history, bot_history)
                        work.add(parent_history)
                        break
                    working_history = parent_history
                except KeyError:
                    # the parent history is not found in the path mapping
                    log.error("Parent history %s is not found", parent_history)
                    break

    @staticmethod
    def _generate_graph(state_histories, hierarchy, symexec_view):
        g = networkx.DiGraph()

        history_to_block = {}

        for state_history in state_histories:
            if state_history not in history_to_block:
                history_to_block[state_history] = QStateBlock(False, symexec_view, history=state_history)
            g.add_node(history_to_block[state_history])

        for src, dst in QPathTree._all_edges_gen(state_histories, hierarchy):
            if src not in history_to_block:
                history_to_block[src] = QStateBlock(False, symexec_view, history=src)
            if dst not in history_to_block:
                history_to_block[dst] = QStateBlock(dst, False, symexec_view, history=dst)
            g.add_edge(history_to_block[src], history_to_block[dst])

        return g

    def _watch_simgr(self, **kwargs) -> None:
        self.reload()



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import sys
import threading
import unittest

import angr
from angr.analyses.decompiler.structured_codegen.c import CFunction, CFunctionCall
from common import start_main_window_and_event_loop, test_location
from PySide6.QtCore import Qt
from PySide6.QtTest import QTest

from angrmanagement.logic.threads import gui_thread_schedule
from angrmanagement.ui.dialogs.rename_label import RenameLabel
from angrmanagement.ui.dialogs.rename_node import RenameNode
from angrmanagement.ui.views import CodeView, DisassemblyView


class TestRenameFunctions(unittest.TestCase):
    def setUp(self):
        self.event = threading.Event()
        _, self.main = start_main_window_and_event_loop(self.event)

    def tearDown(self) -> None:
        self.event.set()
        del self.main

    def _test_rename_a_function_in_disasm_and_pseudocode_views(self):
        main = self.main

        func = main.workspace.main_instance.project.kb.functions["main"]
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        pseudocode_view = main.workspace._get_or_create_view("pseudocode", CodeView)

        # find the node for function
        for _, item in pseudocode_view.codegen.map_pos_to_node.items():
            if isinstance(item.obj, CFunction):
                func_node = item.obj
                break
        else:
            self.fail("The CFunction instance is not found.")

        self.assertEqual(func_node.name, "main")

        # rename the function in the disassembly view
        rlabel = RenameLabel(disasm_view, func.addr, parent=None)
        rlabel._name_box.setText("")
        QTest.keyClicks(rlabel._name_box, "asdf")
        QTest.mouseClick(rlabel._ok_button, Qt.MouseButton.LeftButton)

        self.assertEqual(func.name, "asdf")
        self.assertEqual(func_node.name, "main")

        # rename the function in the pseudocode view
        rnode = RenameNode(code_view=pseudocode_view, node=func_node)
        rnode._name_box.setText("")
        QTest.keyClicks(rnode._name_box, "fdsa")
        QTest.mouseClick(rnode._ok_button, Qt.MouseButton.LeftButton)

        self.assertEqual(func.name, "fdsa")

    def test_rename_a_function_in_disasm_and_pseudocode_views(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        self.assertIsNotNone(func)

        # decompile the function
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        gui_thread_schedule(disasm_view.display_function, args=(func,))
        disasm_view.decompile_current_function()
        main.workspace.job_manager.join_all_jobs()

        # run the jobless method in the GUI thread
        gui_thread_schedule(self._test_rename_a_function_in_disasm_and_pseudocode_views)

    def _test_rename_a_callee_in_pseudocode_view(self):
        main = self.main

        func = main.workspace.main_instance.project.kb.functions["authenticate"]
        _ = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        pseudocode_view = main.workspace._get_or_create_view("pseudocode", CodeView)

        # find the node for function
        for _, item in pseudocode_view.codegen.map_pos_to_node.items():
            if isinstance(item.obj, CFunctionCall) and item.obj.callee_func.name == "authenticate":
                func_node = item.obj
                break
        else:
            self.fail("The CFunction instance is not found.")

        # rename the function in the pseudocode view
        rnode = RenameNode(code_view=pseudocode_view, node=func_node)
        rnode._name_box.setText("")
        QTest.keyClicks(rnode._name_box, "authenticate_1337")
        QTest.mouseClick(rnode._ok_button, Qt.MouseButton.LeftButton)

        self.assertEqual(func.name, "authenticate_1337")

    def test_rename_a_callee_in_pseudocode_view(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        self.assertIsNotNone(func)

        # decompile the function
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        gui_thread_schedule(disasm_view.display_function, args=(func,))
        disasm_view.decompile_current_function()
        main.workspace.job_manager.join_all_jobs()

        # run the jobless method in the GUI thread
        gui_thread_schedule(self._test_rename_a_callee_in_pseudocode_view)


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QProgressDialog

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow


class ProgressDialog(QProgressDialog):
    """ProgressDialog is a dialog that shows the progress of a job on top of the main window."""

    main_window: MainWindow

    def __init__(self, main_window: MainWindow):
        super().__init__("Waiting...", "Cancel", 0, 100, parent=main_window)
        self.main_window = main_window
        self.setAutoClose(False)
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint)
        self.setModal(True)
        self.setMinimumDuration(2**31 - 1)

        self.canceled.connect(self.on_cancel)
        self.close()

    def on_cancel(self) -> None:
        if self.main_window.workspace is None:
            return
        for job in self.main_window.workspace.job_manager.jobs:
            if job.blocking:
                self.main_window.workspace.job_manager.interrupt_current_job()
                break



from __future__ import annotations

import functools
import logging
from collections import defaultdict
from typing import TYPE_CHECKING

from angr.block import Block
from angr.knowledge_plugins.cfg import MemoryData
from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QAction, QCursor
from PySide6.QtWidgets import QApplication, QHBoxLayout, QMenu, QMessageBox, QVBoxLayout

from angrmanagement.data.function_graph import FunctionGraph
from angrmanagement.data.highlight_region import SynchronizedHighlightRegion
from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.commands import ViewCommand
from angrmanagement.logic.disassembly import InfoDock, JumpHistory
from angrmanagement.ui.dialogs.assemble_patch import AssemblePatchDialog
from angrmanagement.ui.dialogs.dependson import DependsOn
from angrmanagement.ui.dialogs.func_doc import FuncDocDialog
from angrmanagement.ui.dialogs.hook import HookDialog
from angrmanagement.ui.dialogs.jumpto import JumpTo
from angrmanagement.ui.dialogs.new_state import NewState
from angrmanagement.ui.dialogs.rename import RenameDialog
from angrmanagement.ui.dialogs.rename_label import RenameLabel
from angrmanagement.ui.dialogs.set_comment import SetComment
from angrmanagement.ui.dialogs.xref import XRefDialog
from angrmanagement.ui.menus.disasm_insn_context_menu import DisasmInsnContextMenu
from angrmanagement.ui.menus.disasm_label_context_menu import DisasmLabelContextMenu
from angrmanagement.ui.views.symexec_view import SymexecView
from angrmanagement.ui.widgets import (
    DisassemblyLevel,
    QAvoidAddrAnnotation,
    QBlockAnnotations,
    QDisasmStatusBar,
    QDisassemblyGraph,
    QFindAddrAnnotation,
    QLinearDisassembly,
    QLinearDisassemblyView,
)
from angrmanagement.ui.widgets.qblock_code import QVariableObj
from angrmanagement.ui.widgets.qinst_annotation import QBreakAnnotation, QHookAnnotation
from angrmanagement.utils import locate_function

from .view import SynchronizedFunctionView

if TYPE_CHECKING:
    import PySide6
    from angr.knowledge_plugins import VariableManager

    from angrmanagement.data.instance import Instance, ObjectContainer
    from angrmanagement.logic.disassembly.info_dock import OperandDescriptor
    from angrmanagement.ui.workspace import Workspace


_l = logging.getLogger(__name__)


class DisassemblyView(SynchronizedFunctionView):
    """
    Disassembly View
    """

    view_visibility_changed = Signal()
    disassembly_level_changed = Signal(DisassemblyLevel)

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("disassembly", workspace, default_docking_position, instance)

        self.base_caption = "Disassembly"
        self._disassembly_level = DisassemblyLevel.MachineCode
        self._show_minimap: bool = True
        self._show_address = True
        self._show_variable = True
        # whether we want to show identifier or not
        self._show_variable_ident = False
        # whether we want to show exception edges and all nodes that are only reachable through exception edges
        self._show_exception_edges = True

        self._prefer_graph = True
        self._current_view: QLinearDisassembly | QDisassemblyGraph | None = None

        self._statusbar = None
        self.jump_history: JumpHistory = JumpHistory()
        self.infodock = InfoDock(self)
        self._variable_recovery_flavor = "fast"
        self.variable_manager: VariableManager | None = None

        self._insn_menu: DisasmInsnContextMenu | None = None
        self._label_menu: DisasmLabelContextMenu | None = None

        self._insn_addr_on_context_menu = None
        self._label_addr_on_context_menu = None

        self._annotation_callbacks = []

        self.width_hint = 800
        self.height_hint = 800

        self._init_widgets()
        self._init_menus()
        self._register_events()

    @classmethod
    def register_commands(cls, workspace: Workspace) -> None:
        """
        Register commands that can be run for this view.
        """
        workspace.command_manager.register_commands(
            [
                ViewCommand("disassembly_view_" + action.__name__, "Disassembly: " + caption, action, cls, workspace)
                for caption, action in [
                    ("Comment", cls.popup_comment_dialog),
                    ("Jump Back", cls.jump_back),
                    ("Jump Forward", cls.jump_forward),
                    ("Jump To", cls.popup_jumpto_dialog),
                    ("Toggle Addresses", cls.toggle_show_address),
                    ("Toggle Exception Edges", cls.toggle_show_exception_edges),
                    ("Toggle Graph/Linear view", cls.toggle_disasm_view),
                    ("Toggle Minimap", cls.toggle_show_minimap),
                    ("Toggle Smart Highlighting", cls.toggle_smart_highlighting),
                    ("Toggle Variable Identifiers", cls.toggle_show_variable_identifier),
                    ("Toggle Variables", cls.toggle_show_variable),
                    ("View AIL", cls.set_disassembly_level_ail),
                    ("View Lifter IR", cls.set_disassembly_level_lifter_ir),
                    ("View Machine Code", cls.set_disassembly_level_machine_code),
                ]
            ]
        )

    @property
    def disassembly_level(self):
        return self._disassembly_level

    def set_disassembly_level(self, level: DisassemblyLevel) -> None:
        self._disassembly_level = level
        self._flow_graph.set_disassembly_level(level)
        self._linear_viewer.set_disassembly_level(level)
        self.disassembly_level_changed.emit(level)
        self.redraw_current_graph()

    def set_disassembly_level_ail(self) -> None:
        self.set_disassembly_level(DisassemblyLevel.AIL)

    def set_disassembly_level_lifter_ir(self) -> None:
        self.set_disassembly_level(DisassemblyLevel.LifterIR)

    def set_disassembly_level_machine_code(self) -> None:
        self.set_disassembly_level(DisassemblyLevel.MachineCode)

    def reload(self) -> None:
        old_infodock = self.infodock.copy()

        self.infodock.initialize()

        # Reload the current graph to make sure it gets the latest information, such as variables.
        self._reload_current_function_if_changed()
        self._current_view.reload(old_infodock=old_infodock)

    def refresh(self) -> None:
        self._current_view.refresh()

    def save_image_to(self, path) -> None:
        if self._flow_graph is not None:
            self._flow_graph.save_image_to(path)

    def setFocus(self) -> None:
        self._flow_graph.setFocus()

    #
    # Properties
    #

    @property
    def disasm(self):
        return self._flow_graph.disasm

    @property
    def show_minimap(self):
        return self._show_minimap

    @property
    def smart_highlighting(self):
        if self._flow_graph is None:
            return False
        if self.infodock is None:
            return False
        return self.infodock.smart_highlighting

    @property
    def show_address(self):
        return self._show_address

    @property
    def show_variable(self):
        return self._show_variable

    @property
    def show_variable_identifier(self):
        return self._show_variable_ident

    @property
    def show_exception_edges(self):
        return self._show_exception_edges

    @property
    def variable_recovery_flavor(self):
        return self._variable_recovery_flavor

    @variable_recovery_flavor.setter
    def variable_recovery_flavor(self, v) -> None:
        if v in ("fast", "accurate") and v != self._variable_recovery_flavor:
            self._variable_recovery_flavor = v
            # TODO: Rerun the variable recovery analysis and update the current view

    @property
    def current_graph(self) -> QLinearDisassemblyView | QDisassemblyGraph:
        """
        Return the current disassembly control, either linear viewer or flow graph.

        :return:    Linear viewer or flow graph.
        """
        return self._current_view

    @property
    def current_function(self) -> ObjectContainer:
        return self.function

    @SynchronizedFunctionView.function.setter
    def function(self, v) -> None:
        if v is not self.function.am_obj:
            self.display_function(v)

    #
    # Callbacks
    #

    # All callbacks are proxies to self.instance. These properties *in this class* may be removed in the near
    # future.
    @property
    def insn_backcolor_callback(self):
        return self.instance.insn_backcolor_callback

    @insn_backcolor_callback.setter
    def insn_backcolor_callback(self, v) -> None:
        self.instance.insn_backcolor_callback = v

    @property
    def label_rename_callback(self):
        return self.instance.label_rename_callback

    @label_rename_callback.setter
    def label_rename_callback(self, v) -> None:
        self.instance.label_rename_callback = v

    @property
    def set_comment_callback(self):
        return self.instance.set_comment_callback

    @set_comment_callback.setter
    def set_comment_callback(self, v) -> None:
        self.instance.set_comment_callback = v

    def on_variable_recovered(self, func_addr: int) -> None:
        if not self.function.am_none and self.function.addr == func_addr:
            self.reload()

    #
    # Events
    #

    def keyPressEvent(self, event) -> None:
        key = event.key()
        if key == Qt.Key.Key_G:
            # jump to window
            self.popup_jumpto_dialog()
            return
        elif key == Qt.Key.Key_Left and QApplication.keyboardModifiers() & Qt.Modifier.ALT != 0:
            # jump back
            self.jump_back()
            return
        elif key == Qt.Key.Key_Right and QApplication.keyboardModifiers() & Qt.Modifier.ALT != 0:
            # jump forward
            self.jump_forward()
            return
        elif key == Qt.Key.Key_A:
            # switch between highlight mode
            self.toggle_smart_highlighting(not self.infodock.smart_highlighting)
            return
        elif key == Qt.Key.Key_Tab:
            # decompile
            self.decompile_current_function()
            return
        elif key == Qt.Key.Key_Semicolon:
            # add comment
            self.popup_comment_dialog()
            return
        elif key == Qt.Key.Key_Space:
            # switch to linear view
            self.toggle_disasm_view()
            event.accept()
            return
        elif key == Qt.Key.Key_Escape:
            # jump back
            self.jump_back()
            return
        elif key == Qt.Key.Key_C:
            self.define_code()
            return
        elif key == Qt.Key.Key_U:
            self.undefine_code()
            return

        super().keyPressEvent(event)

    def redraw_current_graph(self, **kwargs) -> None:  # pylint: disable=unused-argument
        """
        Redraw the graph currently in display.

        :return:    None
        """
        self._current_view.redraw()

    def on_screen_changed(self) -> None:
        self._current_view.refresh()

    def _reload_current_function_if_changed(self) -> None:
        if self._flow_graph.function_graph is not None:
            func_addr = self._flow_graph.function_graph.function.addr

            try:
                func = self.instance.kb.functions.get_by_addr(func_addr)
            except KeyError:
                func = None

            if self._flow_graph.function_graph.function is not func:
                self._display_function(func)

            if func is None:
                self._jump_to(func_addr)

    def _on_cfb_event(self, **kwargs) -> None:
        if not kwargs:
            self._reload_current_function_if_changed()
            self._linear_viewer.reload()

    #
    # UI
    #

    def append_view_menu_actions(self, menu: QMenu) -> None:
        """
        Append a separator and general QActions for this view to a given context menu.
        """
        menu.addSeparator()
        menu.addMenu(self.get_synchronize_with_submenu())

    def contextMenuEvent(self, event: PySide6.QtGui.QContextMenuEvent) -> None:  # pylint: disable=unused-argument
        """
        Display view context menu.
        """
        mnu = QMenu(self)
        mnu.addMenu(self.get_synchronize_with_submenu())
        mnu.exec_(QCursor.pos())

    def instruction_context_menu(self, insn, pos) -> None:
        self._insn_addr_on_context_menu = insn.addr

        # pass in the instruction address
        self._insn_menu.insn_addr = insn.addr
        # pop up the menu
        mnu = self._insn_menu.qmenu(
            extra_entries=list(self.workspace.plugins.build_context_menu_insn(insn)), cached=False
        )
        self.append_view_menu_actions(mnu)
        mnu.exec_(pos)

        self._insn_addr_on_context_menu = None

    def label_context_menu(self, addr: int, pos) -> None:
        self._label_addr_on_context_menu = addr
        self._label_menu.addr = addr
        mnu = self._label_menu.qmenu(cached=False)
        self.append_view_menu_actions(mnu)
        mnu.exec_(pos)
        self._label_addr_on_context_menu = None

    def rename_selected_object(self) -> None:
        """
        Opens dialog for renaming the currently selected QBlockCodeObj
        """
        obj = self.infodock.selected_qblock_code_obj
        if isinstance(obj, QVariableObj):
            dlg = RenameDialog("Rename Variable", obj.obj.name, self)
            dlg.exec_()
            if dlg.result is not None:
                obj.obj.name = dlg.result
                self._current_view.refresh()

    def define_code(self) -> None:
        """
        Redefine selected data as code
        """
        if self.infodock.selected_labels:
            self.workspace.define_code(next(iter(self.infodock.selected_labels)))

    def undefine_code(self) -> None:
        """
        Undefine selected instruction as code, mark it as data
        """
        if self.infodock.selected_insns:
            self.workspace.undefine_code(next(iter(self.infodock.selected_insns)))

    def get_context_menu_for_selected_object(self) -> QMenu | None:
        """
        Returns a QMenu object for the currently selected QBlockCodeObj
        """
        obj = self.infodock.selected_qblock_code_obj
        if isinstance(obj, QVariableObj):
            rename_act = QAction("Re&name", self)
            rename_act.triggered.connect(self.rename_selected_object)
            mnu = QMenu()
            mnu.addActions([rename_act])
            return mnu
        else:
            return None

    def show_context_menu_for_selected_object(self) -> None:
        """
        Spawns a context menu for the currently selected QBlockCodeObj
        """
        mnu = self.get_context_menu_for_selected_object()
        if mnu is not None:
            self.append_view_menu_actions(mnu)
            mnu.exec_(QCursor.pos())

    def popup_jumpto_dialog(self) -> None:
        JumpTo(self, parent=self).exec_()

    def popup_rename_label_dialog(self) -> None:
        label_addr_tpl = self._address_in_selection()
        if label_addr_tpl is None:
            return

        type_, label_addr = label_addr_tpl
        dialog = RenameLabel(self, label_addr, parent=self, full_refresh=type_ == "operand")
        dialog.exec_()

    def popup_comment_dialog(self) -> None:
        comment_addr = self._instruction_address_in_selection()
        if comment_addr is None:
            return

        dialog = SetComment(self.workspace, comment_addr, parent=self)
        dialog.exec_()

    def popup_newstate_dialog(self, async_: bool = True) -> None:
        addr = self._instruction_address_in_selection()
        if addr is None:
            return

        dialog = NewState(self.workspace, self.instance, addr=addr, create_simgr=True, parent=self)
        if async_:
            dialog.show()
        else:
            dialog.exec_()

    def popup_hook_dialog(self, async_: bool = True, addr: int | None = None) -> None:
        addr = addr or self._instruction_address_in_selection()

        if addr is None:
            return

        dialog = HookDialog(self.workspace, addr=addr, parent=self)
        if async_:
            dialog.show()
        else:
            dialog.exec_()

    def popup_func_doc_dialog(self, instr_addr) -> None:
        """
        Spawns a popup dialog for the currently selected call instruction func_docs
        """
        if self._flow_graph is None:
            return
        block = self._flow_graph._insaddr_to_block.get(instr_addr, None)
        if block:
            instr = block.addr_to_insns[instr_addr]
            if instr is None or instr.insn.type != "call":
                return
            out_targets = instr.out_branch.targets
            if len(out_targets) != 1:
                return

            target = next(iter(out_targets))
            operand = instr.get_operand(0)

            doc_tuple = GlobalInfo.library_docs.get_docstring_for_func_name(operand.text)
            if doc_tuple is None:
                doc_string = f"Cannot find local documentation for function {operand.text}."
                url = "http://"
                ftype = "<>"
                doc_tuple = (doc_string, url, ftype)
            dialog = FuncDocDialog(self.instance, addr=target, name=operand.text, doc_tuple=doc_tuple, parent=self)
            dialog.show()

    def popup_dependson_dialog(self, addr: int | None = None, use_operand: bool = False, func: bool = False):
        if use_operand:
            r = self._flow_graph.get_selected_operand_info()
            if r is not None:
                _, addr, operand = r
            else:
                QMessageBox.critical(
                    self,
                    "No operand",
                    "Please select an operand first.",
                    buttons=QMessageBox.StandardButton.Ok,
                )
                return
        else:
            if addr is None:
                raise ValueError("No address is provided.")  # this is a programming error
            operand = None

        # if a function target is selected, switch to function mode
        if operand is not None and not func and operand._branch_target is not None and operand._is_target_func:
            func = True
            addr = operand._branch_target

        if func:
            # attempt to pass in a function
            try:
                the_func = self.instance.kb.functions.get_by_addr(addr)
            except KeyError:
                the_func = None
        else:
            the_func = None

        dependson = DependsOn(addr, operand, func=the_func, parent=self)
        dependson.exec_()

        if dependson.location is not None and dependson.arg is not None:
            # track function argument
            self.workspace._main_window.run_dependency_analysis(
                func_addr=addr,
                func_arg_idx=dependson.arg,
            )

    def parse_operand_and_popup_xref_dialog(self, ins_addr, operand, async_: bool = True) -> None:
        if operand is not None:
            if operand.variable is not None:
                # Display cross references to this variable
                self.popup_xref_dialog(addr=ins_addr, variable=operand.variable, async_=async_)
            elif operand.is_constant:
                # Display cross references to an address
                self.popup_xref_dialog(addr=ins_addr, dst_addr=operand.constant_value, async_=async_)
            elif operand.is_constant_memory:
                # Display cross references to an address
                self.popup_xref_dialog(addr=ins_addr, dst_addr=operand.constant_memory_value, async_=async_)

    def popup_xref_dialog(self, addr: int | None = None, variable=None, dst_addr=None, async_: bool = True) -> None:
        if variable is not None:
            dialog = XRefDialog(
                addr=addr,
                variable_manager=self.variable_manager,
                variable=variable,
                instance=self.instance,
                disassembly_view=self,
                parent=self,
            )
        else:
            dialog = XRefDialog(
                addr=addr,
                xrefs_manager=self.instance.project.kb.xrefs,
                dst_addr=dst_addr,
                instance=self.instance,
                disassembly_view=self,
                parent=self,
            )
        if async_:
            dialog.show()
        else:
            dialog.exec_()

    def popup_patch_dialog(self) -> None:
        dlg = AssemblePatchDialog(self._insn_addr_on_context_menu, self.instance)
        dlg.exec_()

    #
    # Public methods
    #

    def toggle_disasm_view(self, prefer: bool = True) -> None:
        if self._flow_graph.isHidden():
            # Show flow graph
            self.display_disasm_graph(prefer)
        else:
            # Show linear viewer
            self.display_linear_viewer(prefer)

    def display_disasm_graph(self, prefer: bool = True) -> None:
        if prefer:
            self._prefer_graph = True

        self._linear_viewer.hide()
        self._current_view = self._flow_graph
        self._flow_graph.show()

        if self.infodock.selected_insns:
            # display the currently selected instruction
            self._jump_to(next(iter(self.infodock.selected_insns)))
        elif self.function.am_obj is not None:
            self._flow_graph.show_instruction(self.function.addr)

        self._flow_graph.setFocus()
        self.view_visibility_changed.emit()
        self._flow_graph.refresh()

    def display_linear_viewer(self, prefer: bool = True) -> None:
        if prefer:
            self._prefer_graph = False

        self._flow_graph.hide()
        self._current_view = self._linear_viewer
        self._linear_viewer.show()

        if self.infodock.selected_insns:
            # display the currently selected instruction
            self._linear_viewer.show_instruction(next(iter(self.infodock.selected_insns)))
        elif self.function.am_obj is not None:
            self._linear_viewer.show_instruction(self.function.addr)

        self._linear_viewer.setFocus()
        self.view_visibility_changed.emit()
        self._linear_viewer.refresh()

    def display_function(self, function) -> None:
        if function is None:
            return
        if (
            function.addr not in self.instance.kb.variables.function_managers
            and self.instance.variable_recovery_job is not None
        ):
            # variable information is not available
            # prioritize the analysis of this function
            self.instance.variable_recovery_job.prioritize_function(function.addr)
        self.jump_history.jump_to(function.addr)
        self._display_function(function)

    def decompile_current_function(self) -> None:
        if self.function.am_obj is not None:
            try:
                curr_ins = next(iter(self.infodock.selected_insns))
            except StopIteration:
                curr_ins = None

            self.workspace.decompile_function(self.function.am_obj, curr_ins=curr_ins)

    def toggle_show_minimap(self, show_minimap: bool | None = None) -> None:
        """
        Toggle minimap display preference
        """
        if show_minimap is None:
            show_minimap = not self._show_minimap
        self._show_minimap = show_minimap
        self._current_view.refresh()

    def toggle_smart_highlighting(self, enabled: bool | None = None) -> None:
        """
        Toggle between the smart highlighting mode and the text-based highlighting mode.
        """
        if enabled is None:
            enabled = not self.infodock.smart_highlighting
        self.infodock.smart_highlighting = enabled
        self._flow_graph.refresh()
        self._linear_viewer.refresh()

    def toggle_show_address(self, show_address: bool | None = None) -> None:
        """
        Toggle whether addresses are shown on disassembly graph.
        """
        if show_address is None:
            show_address = not self._show_address
        self._show_address = show_address
        self._current_view.refresh()

    def toggle_show_variable(self, show_variable: bool | None = None) -> None:
        """
        Toggle whether variables are shown on disassembly graph.
        """
        if show_variable is None:
            show_variable = not self._show_variable
        self._show_variable = show_variable
        self._current_view.refresh()

    def toggle_show_variable_identifier(self, show_ident: bool | None = None) -> None:
        """
        Toggle whether variable identifiers are shown on disassembly graph.
        """
        if show_ident is None:
            show_ident = not self._show_variable_ident
        self._show_variable_ident = show_ident
        self._current_view.refresh()

    def toggle_show_exception_edges(self, show_exception_edges: bool | None = None) -> None:
        """
        Toggle whether exception edges and the nodes that are only reachable through exception edges should be shown
        or not.

        :param bool show_exception_edges:   Whether exception edges should be shown or not.
        :return:                            None
        """

        if show_exception_edges != self._show_exception_edges:
            self._show_exception_edges = show_exception_edges

            # reset the function graph
            if self._flow_graph.function_graph is not None:
                self._flow_graph.function_graph.exception_edges = show_exception_edges
                self._flow_graph.function_graph.clear_cache()
                self._flow_graph.reload()

    def jump_to(self, addr: int, src_ins_addr=None, use_animation: bool = False) -> bool:
        # Record the current instruction address first
        if src_ins_addr is not None:
            self.jump_history.record_address(src_ins_addr)

        self.jump_history.jump_to(addr)
        self._jump_to(addr, use_animation=use_animation)

        return True

    def jump_back(self) -> None:
        addr = self.jump_history.backtrack()
        if addr is not None:
            self._jump_to(addr, use_animation=False)

    def jump_forward(self) -> None:
        addr = self.jump_history.forwardstep()
        if addr is not None:
            self._jump_to(addr, use_animation=False)

    def jump_to_history_position(self, pos: int) -> None:
        addr = self.jump_history.step_position(pos)
        if addr is not None:
            self._jump_to(addr, use_animation=False)

    def select_label(self, label_addr) -> None:
        self.infodock.select_label(label_addr)

    def rename_label(self, addr: int, new_name: str, is_func: bool = False, full_refresh: bool = False) -> None:
        if self._flow_graph.disasm is not None:
            is_renaming = False

            kb = self._flow_graph.disasm.kb
            if is_func:
                func = kb.functions.get_by_addr(addr)
                is_renaming = True
                if new_name:
                    func.name = new_name
                else:
                    # restore to the default name
                    func.name = f"sub_{addr:x}"
            else:
                if new_name == "":
                    if addr in kb.labels:
                        del kb.labels[addr]
                else:
                    if addr in kb.labels:
                        is_renaming = True
                    kb.labels[addr] = new_name

            # callback first
            if self.instance.label_rename_callback:
                self.instance.label_rename_callback(addr=addr, new_name=new_name)

            if full_refresh:
                # redraw the entire graph. required if a data address is renamed.
                self._flow_graph.refresh()
            else:
                # redraw the current block
                self._flow_graph.update_label(addr, is_renaming=is_renaming)

    def avoid_addr_in_exec(self, addr: int) -> None:
        self.workspace._get_or_create_view("symexec", SymexecView).avoid_addr_in_exec(addr)

    def find_addr_in_exec(self, addr: int) -> None:
        self.workspace._get_or_create_view("symexec", SymexecView).find_addr_in_exec(addr)

    def run_induction_variable_analysis(self) -> None:
        if self._flow_graph.induction_variable_analysis:
            self._flow_graph.induction_variable_analysis = None
        else:
            analyses = self.instance.project.analyses
            ana = analyses.AffineRelationAnalysis(self._flow_graph._function_graph.function)
            self._flow_graph.induction_variable_analysis = ana
        self._flow_graph.refresh()

    def fetch_qblock_annotations(self, qblock):
        addr_to_annotations = defaultdict(list)
        for annotations_ in self.workspace.plugins.build_qblock_annotations(qblock):
            addr_to_annotations[annotations_.addr].append(annotations)
        for addr in qblock.addr_to_insns:
            if addr in self.instance.project._sim_procedures:
                hook_annotation = QHookAnnotation(addr)
                addr_to_annotations[addr].append(hook_annotation)
            view = self.workspace.view_manager.first_view_in_category("symexec")
            if view is not None:
                qsimgrs = view._simgrs
                if addr in qsimgrs.find_addrs:
                    addr_to_annotations[addr].append(QFindAddrAnnotation(addr, qsimgrs))
                if addr in qsimgrs.avoid_addrs:
                    addr_to_annotations[addr].append(QAvoidAddrAnnotation(addr, qsimgrs))
                for bp in self.instance.breakpoint_mgr.get_breakpoints_at(addr):
                    addr_to_annotations[addr].append(QBreakAnnotation(bp))

        return QBlockAnnotations(addr_to_annotations, parent=qblock, disasm_view=self)

    def update_highlight_regions_for_synchronized_views(self, **kwargs) -> None:  # pylint: disable=unused-argument
        """
        Highlight each selected instruction in synchronized views.
        """
        regions = []
        for addr in self.infodock.selected_insns:
            s = self._get_instruction_size(addr)
            if s is not None:
                regions.append(SynchronizedHighlightRegion(addr, s))
        self.set_synchronized_highlight_regions(regions)

    #
    # Initialization
    #

    def _init_widgets(self) -> None:
        self._linear_viewer = QLinearDisassembly(self.instance, self, parent=self)
        self._flow_graph = QDisassemblyGraph(self.instance, self, parent=self)
        self._statusbar = QDisasmStatusBar(self, parent=self)

        vlayout = QVBoxLayout()
        vlayout.addWidget(self._statusbar)
        vlayout.addWidget(self._flow_graph)
        vlayout.addWidget(self._linear_viewer)
        vlayout.setSpacing(0)
        vlayout.setContentsMargins(0, 0, 0, 0)

        vlayout.setStretchFactor(self._flow_graph, 1)
        vlayout.setStretchFactor(self._linear_viewer, 1)
        vlayout.setStretchFactor(self._statusbar, 0)

        hlayout = QHBoxLayout()
        hlayout.addLayout(vlayout)
        hlayout.setSpacing(20)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)

        self.display_disasm_graph()

        self.workspace.plugins.instrument_disassembly_view(self)

    def _init_menus(self) -> None:
        self._insn_menu = DisasmInsnContextMenu(self)
        self._label_menu = DisasmLabelContextMenu(self)

    def _register_events(self) -> None:
        # redraw the current graph if instruction/operand selection changes
        self.infodock.selected_insns.am_subscribe(self.redraw_current_graph)
        self.infodock.selected_insns.am_subscribe(self.update_highlight_regions_for_synchronized_views)
        self.infodock.selected_operands.am_subscribe(self.redraw_current_graph)
        self.infodock.selected_blocks.am_subscribe(self.redraw_current_graph)
        self.infodock.hovered_block.am_subscribe(self.redraw_current_graph)
        self.infodock.hovered_edge.am_subscribe(self.redraw_current_graph)
        self.infodock.selected_labels.am_subscribe(self.redraw_current_graph)
        self.infodock.selected_variables.am_subscribe(self.redraw_current_graph)
        self.infodock.qblock_code_obj_selection_changed.connect(self.redraw_current_graph)
        self.workspace.current_screen.am_subscribe(self.on_screen_changed)
        self.instance.breakpoint_mgr.breakpoints.am_subscribe(self._on_breakpoints_updated)
        self.instance.cfb.am_subscribe(self._on_cfb_event)
        self.function.am_subscribe(functools.partial(self.display_function, self.function.am_obj))

    def _on_breakpoints_updated(self, **kwargs) -> None:  # pylint:disable=unused-argument
        self.refresh()

    def _unregister_events(self) -> None:
        self.infodock.selected_insns.am_unsubscribe(self.redraw_current_graph)
        self.infodock.selected_insns.am_unsubscribe(self.update_highlight_regions_for_synchronized_views)
        self.infodock.selected_operands.am_unsubscribe(self.redraw_current_graph)
        self.infodock.selected_blocks.am_unsubscribe(self.redraw_current_graph)
        self.infodock.hovered_block.am_unsubscribe(self.redraw_current_graph)
        self.infodock.hovered_edge.am_unsubscribe(self.redraw_current_graph)
        self.infodock.selected_labels.am_unsubscribe(self.redraw_current_graph)
        self.infodock.selected_variables.am_unsubscribe(self.redraw_current_graph)
        self.infodock.qblock_code_obj_selection_changed.disconnect(self.redraw_current_graph)
        self.workspace.current_screen.am_unsubscribe(self.on_screen_changed)
        self.instance.breakpoint_mgr.breakpoints.am_unsubscribe(self._on_breakpoints_updated)
        self.instance.cfb.am_unsubscribe(self._on_cfb_event)

    def closeEvent(self, event) -> None:
        self._unregister_events()
        super().closeEvent(event)

    #
    # Private methods
    #

    def _display_function(self, the_func) -> None:
        if the_func is not None:
            self.set_synchronized_cursor_address(the_func.addr)

        self.function.am_obj = the_func
        self.function.am_event()

        # set status bar
        self._statusbar.function = the_func

        # variable recovery
        variable_manager = self.instance.project.kb.variables
        self.variable_manager = variable_manager
        self.infodock.variable_manager = variable_manager

        # clear existing selected instructions and operands
        self.infodock.clear_selection()

        if self._flow_graph.function_graph is None or self._flow_graph.function_graph.function is not the_func:
            self._flow_graph.function_graph = (
                None
                if the_func is None
                else FunctionGraph(
                    function=the_func,
                    exception_edges=self.show_exception_edges,
                )
            )

        if self._current_view is self._linear_viewer and the_func is not None:
            self._linear_viewer.navigate_to_addr(the_func.addr)

        console_view = self.workspace.view_manager.first_view_in_category("console")
        if console_view is not None:
            console_view.set_current_function(the_func)

    def _jump_to(self, addr: int, use_animation: bool = False) -> bool:
        if self._prefer_graph and self._current_view is self._linear_viewer:
            self.display_disasm_graph(prefer=False)

        if self._current_view is not self._linear_viewer:
            function = locate_function(self.instance, addr)
            if function is not None:
                self._display_function(function)
                instr_addr = function.addr_to_instruction_addr(addr)
                if instr_addr is None:
                    instr_addr = addr
                self.infodock.select_instruction(instr_addr, unique=True, use_animation=use_animation)
                return True

            # it does not belong to any function - we need to switch to linear view mode
            self.display_linear_viewer(prefer=False)

        try:
            item = self.instance.cfb.floor_item(addr)
            _, item = item
            if isinstance(item, MemoryData) and addr < (item.addr + item.size):
                self.infodock.select_label(item.addr)
            elif isinstance(item, Block) and item.size and addr < (item.addr + item.size):
                addr = max(a for a in item.instruction_addrs if a <= addr)
                self.infodock.select_instruction(addr, unique=True, use_animation=use_animation)
                return True  # select_instruction will navigate
        except KeyError:
            pass

        self._linear_viewer.navigate_to_addr(addr)
        return True

    #
    # Utils
    #

    def _address_in_selection(self) -> tuple[str, int] | None:
        if self._insn_addr_on_context_menu is not None:
            return "insn", self._insn_addr_on_context_menu
        if len(self.infodock.selected_operands) == 1:
            selected_operand: OperandDescriptor = next(iter(self.infodock.selected_operands.values()))
            if selected_operand.num_value is not None:
                return "operand", selected_operand.num_value
        if len(self.infodock.selected_insns) == 1:
            return "insn", next(iter(self.infodock.selected_insns))
        if len(self.infodock.selected_labels) == 1:
            return "insn", next(iter(self.infodock.selected_labels))
        return None

    def _instruction_address_in_selection(self) -> int | None:
        if self._insn_addr_on_context_menu is not None:
            return self._insn_addr_on_context_menu
        if len(self.infodock.selected_insns) == 1:
            return next(iter(self.infodock.selected_insns))
        if len(self.infodock.selected_labels) == 1:
            return next(iter(self.infodock.selected_labels))
        return None

    def _get_instruction_size(self, addr: int) -> int | None:
        kb = self.instance.project.kb
        f = kb.functions.floor_func(addr)
        if f is None:
            return None
        return f.instruction_size(addr)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFrame,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QSpinBox,
    QSplitter,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.config import IMG_LOCATION
from angrmanagement.data.analysis_options import (
    AnalysesConfiguration,
    AnalysisOption,
    BoolAnalysisOption,
    ChoiceAnalysisOption,
    IntAnalysisOption,
    StringAnalysisOption,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from angrmanagement.ui.workspace import Workspace


class AnalysisOptionWidgetMapper:
    """
    Analysis option widget creation and event handling.
    """

    option: AnalysisOption
    widget: QWidget | None

    def __init__(self, option: AnalysisOption) -> None:
        self.option = option
        self.widget = None

    def create_widget(self) -> QWidget:
        raise NotImplementedError

    @classmethod
    def get_mapper_for_option(cls, option: AnalysisOption) -> AnalysisOptionWidgetMapper:
        if isinstance(option, BoolAnalysisOption):
            return BoolAnalysisOptionWidgetMapper(option)
        elif isinstance(option, IntAnalysisOption):
            return IntAnalysisOptionWidgetMapper(option)
        elif isinstance(option, ChoiceAnalysisOption):
            return ChoiceAnalysisOptionWidgetMapper(option)
        elif isinstance(option, StringAnalysisOption):
            return StringAnalysisOptionWidgetMapper(option)
        else:
            raise ValueError("Mapper not implemented")


class BoolAnalysisOptionWidgetMapper(AnalysisOptionWidgetMapper):
    """
    Analysis option widget creation and event handling for boolean options.
    """

    option: BoolAnalysisOption

    def create_widget(self, parent=None) -> QCheckBox:
        self.widget = QCheckBox(parent)
        self.widget.setText(self.option.display_name)
        if self.option.tooltip:
            self.widget.setToolTip(self.option.tooltip)
        self.widget.setChecked(self.option.value)
        self.widget.stateChanged.connect(self._on_checkbox_changed)
        return self.widget

    def _on_checkbox_changed(self, _) -> None:
        self.option.value = self.widget.isChecked()


class StringAnalysisOptionWidgetMapper(AnalysisOptionWidgetMapper):
    """
    Analysis option widget for string answers
    """

    option: StringAnalysisOption

    def __init__(self, option: AnalysisOption) -> None:
        super().__init__(option)

        self.checkbox = None
        self.textbox = None

    def create_widget(self, parent=None) -> QWidget:
        self.textbox = QLineEdit(self.option.value)
        self.textbox.textChanged.connect(self._on_text_changed)

        layout = QHBoxLayout()

        if self.option.optional:
            checkbox = QCheckBox()
            checkbox.setText(self.option.display_name)
            if self.option.tooltip:
                checkbox.setToolTip(self.option.tooltip)
            checkbox.clicked.connect(self._on_toggled)
            self.checkbox = checkbox
            layout.addWidget(checkbox)
            self._on_toggled()  # enable or disable the textbox
        else:
            lbl = QLabel()
            lbl.setText(self.option.display_name)
            if self.option.tooltip:
                lbl.setToolTip(self.option.tooltip)
            layout.addWidget(lbl)

        layout.addWidget(self.textbox)
        self.widget = QWidget(parent)
        self.widget.setLayout(layout)
        return self.widget

    def _on_toggled(self) -> None:
        self.textbox.setEnabled(self.checkbox.isChecked())
        self.option.enabled = self.checkbox.isChecked()

    def _on_text_changed(self, value: str) -> None:
        self.option.value = value


class IntAnalysisOptionWidgetMapper(AnalysisOptionWidgetMapper):
    """
    Analysis option widget creation and event handling for integer options.
    """

    option: IntAnalysisOption

    def create_widget(self, parent=None) -> QWidget:
        spinbox = QSpinBox()
        spinbox.setValue(self.option.value)
        if self.option.tooltip:
            spinbox.setToolTip(self.option.tooltip)
        spinbox.valueChanged.connect(self._on_dial_changed)
        if self.option.minimum_value is not None:
            spinbox.setMinimum(self.option.minimum_value)
        if self.option.maximum_value is not None:
            spinbox.setMaximum(self.option.maximum_value)

        lbl = QLabel()
        lbl.setText(self.option.display_name)
        if self.option.tooltip:
            lbl.setToolTip(self.option.tooltip)

        layout = QHBoxLayout()
        layout.addWidget(lbl)
        layout.addWidget(spinbox)

        self.widget = QWidget(parent)
        self.widget.setLayout(layout)
        return self.widget

    def _on_dial_changed(self, value: int) -> None:
        self.option.value = value


class ChoiceAnalysisOptionWidgetMapper(AnalysisOptionWidgetMapper):
    """
    Analysis option widget creation and event handling for choice options.
    """

    option: ChoiceAnalysisOption

    def __init__(self, option: ChoiceAnalysisOption) -> None:
        super().__init__(option)
        self.combobox = None

    def create_widget(self, parent=None) -> QWidget:
        self.combobox = QComboBox()
        for data, txt in self.option.choices.items():
            self.combobox.addItem(txt, data)
        self.combobox.setCurrentIndex(self.combobox.findData(self.option.value))
        self.combobox.currentIndexChanged.connect(self._on_combo_changed)

        lbl = QLabel()
        lbl.setText(self.option.display_name)
        if self.option.tooltip:
            lbl.setToolTip(self.option.tooltip)

        layout = QHBoxLayout()
        layout.addWidget(lbl)
        layout.addStretch()
        layout.addWidget(self.combobox)

        self.widget = QWidget(parent)
        self.widget.setLayout(layout)
        return self.widget

    def _on_combo_changed(self, index) -> None:
        self.option.value = self.combobox.itemData(index)


class AnalysisOptionsDialog(QDialog):
    """
    Dialog displaying available analyses and configuration options.
    """

    def __init__(self, analyses: AnalysesConfiguration, workspace: Workspace, parent=None) -> None:
        super().__init__(parent)
        self._workspace: Workspace = workspace
        self._analyses: AnalysesConfiguration = analyses
        self._mappers: Sequence[AnalysisOptionWidgetMapper] = []
        self._init_widgets()
        self.setWindowTitle("Run Analysis")
        self.setMinimumSize(self.sizeHint())
        self.adjustSize()

    def sizeHint(self):  # pylint: disable=no-self-use
        return QSize(800, 600)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)

        #
        # Available analyses
        #
        self._analysis_list = QListWidget()
        layout = QVBoxLayout()
        layout.addWidget(self._analysis_list)
        analyses_gbox = QGroupBox("Available Analyses")
        analyses_gbox.setLayout(layout)

        for analysis in self._analyses.analyses:
            item = QListWidgetItem(analysis.display_name, self._analysis_list)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(Qt.CheckState.Checked if analysis.enabled else Qt.CheckState.Unchecked)
            self._analysis_list.addItem(item)

        #
        # Analysis details
        #
        self._analysis_description_label = QLabel()
        self._analysis_description_label.setWordWrap(True)
        layout = QVBoxLayout()
        layout.addWidget(self._analysis_description_label)
        description_gbox = QGroupBox("Description")
        description_gbox.setLayout(layout)

        self._options_layout = QVBoxLayout()
        options_gbox = QGroupBox("Options")
        options_gbox.setLayout(self._options_layout)

        details_layout = QVBoxLayout()
        details_layout.addWidget(description_gbox)
        details_layout.addWidget(options_gbox)

        splitter = QSplitter()
        left_frame = QFrame()
        left_layout = QVBoxLayout()
        left_layout.addWidget(analyses_gbox)
        left_frame.setLayout(left_layout)
        splitter.addWidget(left_frame)
        right_frame = QFrame()
        right_frame.setLayout(details_layout)
        splitter.addWidget(right_frame)
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)

        left_margins = left_layout.contentsMargins()
        left_layout.setContentsMargins(0, 0, left_margins.right(), 0)
        details_margins = details_layout.contentsMargins()
        details_layout.setContentsMargins(details_margins.left(), 0, 0, 0)
        self.main_layout.addWidget(splitter)

        #
        # Dialog buttons
        #
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        ok_button = buttons.button(QDialogButtonBox.StandardButton.Ok)
        ok_button.setText("&Run Analysis")
        ok_button.setIcon(QIcon(os.path.join(IMG_LOCATION, "run-icon.svg")))
        buttons.accepted.connect(self._on_run_clicked)
        buttons.rejected.connect(self.reject)
        self.main_layout.addWidget(buttons)

        self._analysis_list.itemSelectionChanged.connect(self._update_item_details)
        self._analysis_list.itemChanged.connect(self._on_item_changed)

        if len(self._analyses) > 0:
            self._analysis_list.setCurrentRow(0)

        ok_button.setFocus()

    def _update_item_details(self) -> None:
        while self._options_layout.count():
            item = self._options_layout.takeAt(0)
            widget = item.widget()
            if widget:
                widget.deleteLater()
        self._mappers = []

        selected = self._analysis_list.selectedItems()
        if selected:
            analysis = self._analyses[self._analysis_list.indexFromItem(selected[0]).row()]
            self._analysis_description_label.setText(analysis.description)

            for option in analysis.options.values():
                mapper = AnalysisOptionWidgetMapper.get_mapper_for_option(option)
                widget = mapper.create_widget(self)
                self._options_layout.addWidget(widget)
                self._mappers.append(mapper)
        else:
            self._analysis_description_label.setText("Select analysis to view options.")

        self._options_layout.addStretch()

    #
    # Event handlers
    #

    def _on_item_changed(self, item) -> None:
        analysis = self._analyses[self._analysis_list.indexFromItem(item).row()]
        analysis.enabled = item.checkState() == Qt.CheckState.Checked
        self._update_item_details()

    def _on_run_clicked(self) -> None:
        self.accept()



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.calling_conventions import SimRegArg
from PySide6.QtCore import QRectF, Qt
from PySide6.QtGui import QCursor, QPainter
from PySide6.QtWidgets import QApplication, QGraphicsSimpleTextItem

from angrmanagement.config import Conf
from angrmanagement.utils.func import type2str

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angr.sim_type import SimTypeFunction

    from angrmanagement.ui.workspace import Workspace


class QFunctionHeader(QCachedGraphicsItem):
    def __init__(
        self, addr: int, name: str, prototype, args, config, disasm_view, workspace: Workspace, infodock, parent=None
    ) -> None:
        super().__init__(parent=parent)

        self.workspace = workspace
        self.addr = addr
        self.name = name
        self.prototype: SimTypeFunction = prototype
        self.args = args
        self.infodock = infodock

        self._width = 0
        self._height = 0

        self._config = config
        self._disasm_view = disasm_view

        self._return_type_width = None
        self._arg_str_list = None
        self._args_str = None

        self._function_name_item: QGraphicsSimpleTextItem = None
        self._args_str_item: QGraphicsSimpleTextItem = None
        self._prototype_arg_item: QGraphicsSimpleTextItem | None = None

        self._init_widgets()

    def refresh(self) -> None:
        pass

    def paint(self, painter, option, widget) -> None:
        painter.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)

        if self.infodock.is_label_selected(self.addr):
            highlight_color = Conf.disasm_view_label_highlight_color
            painter.setBrush(highlight_color)
            painter.setPen(highlight_color)
            painter.drawRect(0, 0, self.width, self.height)

    #
    # Event handlers
    #

    def mousePressEvent(self, event) -> None:
        if event.button() == Qt.MouseButton.LeftButton:
            self.infodock.select_label(self.addr)
        elif (
            event.button() == Qt.MouseButton.RightButton
            and QApplication.keyboardModifiers() == Qt.KeyboardModifier.NoModifier
        ):
            if self.addr not in self.infodock.selected_labels:
                self.infodock.select_label(self.addr)
            self._disasm_view.label_context_menu(self.addr, QCursor.pos())

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        if self.args is not None:
            self._arg_str_list = []
            for arg in self.args:
                if isinstance(arg, SimRegArg):
                    self._arg_str_list.append(arg.reg_name)
                else:
                    self._arg_str_list.append(str(arg))

            self._args_str = "Args: ({})".format(", ".join(self._arg_str_list))
        else:
            self._args_str = ""

        #
        # prototype
        #

        if self.prototype is None:
            # Just print the function name
            self._function_name_item = QGraphicsSimpleTextItem(self.name, self)
            self._function_name_item.setFont(self._config.code_font)
            self._function_name_item.setBrush(self._config.disasm_view_function_color)

        else:
            # print the prototype

            proto_str = ""

            # Type of the return value
            rt = type2str(self.prototype.returnty)
            proto_str += rt

            # space
            proto_str += " "

            # function name
            proto_str += self.name

            # left parenthesis
            proto_str += "("

            # arguments
            for i, arg_type in enumerate(self.prototype.args):
                type_str = type2str(arg_type)
                proto_str += type_str + " "

                if self.prototype.arg_names and i < len(self.prototype.arg_names):
                    arg_name = self.prototype.arg_names[i]
                else:
                    arg_name = "arg_%d" % i
                proto_str += arg_name

                if i < len(self.prototype.args) - 1:
                    # splitter
                    proto_str += ", "

            # right parenthesis
            proto_str += ")"

            self._prototype_arg_item = QGraphicsSimpleTextItem(proto_str, self)
            self._prototype_arg_item.setFont(self._config.code_font)
            self._prototype_arg_item.setBrush(self._config.disasm_view_function_color)

        # arguments
        if self._arg_str_list is not None:
            s = "Args: (" + ", ".join(self._arg_str_list) + ")"
            self._args_str_item = QGraphicsSimpleTextItem(s, self)
            self._args_str_item.setFont(self._config.code_font)
            self._args_str_item.setBrush(self._config.disasm_view_function_color)

        self._layout_items_and_update_size()

    def _layout_items_and_update_size(self) -> None:
        x, y = 0, 0

        if self._function_name_item is not None:
            # function anme
            self._function_name_item.setPos(x, y)
            x += self._function_name_item.boundingRect().width()
            height = self._function_name_item.boundingRect().height()
        elif self._prototype_arg_item is not None:
            # prototype
            self._prototype_arg_item.setPos(x, y)
            x += self._prototype_arg_item.boundingRect().width()
            height = self._prototype_arg_item.boundingRect().height()
        else:
            height = 0

        # new line
        max_x = x
        x = 0
        y += height

        # arguments
        if self._args_str_item is not None:
            self._args_str_item.setPos(x, y)
            x += self._args_str_item.boundingRect().width()
            y += self._args_str_item.boundingRect().height()

        max_x = max(x, max_x)
        self._width = max_x
        self._height = y
        self.recalculate_size()

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtGui import QAction, QKeySequence

from angrmanagement.ui.icons import icon

from .menu import Menu, MenuEntry, MenuSeparator

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow


class HelpMenu(Menu):
    """
    Main 'Help' menu
    """

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__("&Help", parent=main_window)

        self.entries.extend(
            [
                MenuEntry(
                    "&Documentation", main_window.open_doc_link, shortcut=QKeySequence("Alt+H"), icon=icon("docs")
                ),
                MenuSeparator(),
                MenuEntry("About angr...", main_window.open_about_dialog, role=QAction.MenuRole.AboutRole),
            ]
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from angrmanagement.logic import GlobalInfo

from .menu import Menu, MenuEntry

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class FunctionContextMenu(Menu):
    def __init__(self, workspace: Workspace, parent) -> None:
        super().__init__("Function", parent=parent)
        self.workspace = workspace

        self.funcs = []

        # TODO add Rename, Change Type, xrefs, etc

    def set(self, funcs):
        self.funcs = funcs
        return self

    def qmenu(self, extra_entries=None):
        self.entries = []
        if len(self.funcs):
            self.entries.append(
                MenuEntry("Show Function Info", lambda: self.workspace.show_function_info(self.funcs[0]))
            )
        if extra_entries is None:
            extra_entries = ()
        return super().qmenu(
            extra_entries=list(GlobalInfo.main_window.workspace.plugins.build_context_menu_functions(self.funcs))
            + list(extra_entries),
            cached=False,
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from angr.analyses.decompiler.structured_codegen import DummyStructuredCodeGenerator
from angr.analyses.decompiler.structured_codegen.c import CConstant, CFunctionCall, CStructuredCodeGenerator, CVariable
from angr.knowledge_plugins.functions.function import Function
from angr.sim_variable import SimMemoryVariable
from PySide6.QtCore import Qt
from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import (
    QComboBox,
    QDockWidget,
    QFrame,
    QHBoxLayout,
    QMainWindow,
    QTextEdit,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.config import Conf
from angrmanagement.data.jobs import DecompileFunctionJob, VariableRecoveryJob
from angrmanagement.data.object_container import ObjectContainer
from angrmanagement.logic.disassembly import JumpHistory
from angrmanagement.ui.dialogs.jumpto import JumpTo
from angrmanagement.ui.documents import QCodeDocument
from angrmanagement.ui.toolbars import NavToolbar
from angrmanagement.ui.widgets.qccode_edit import QCCodeEdit
from angrmanagement.ui.widgets.qdecomp_options import QDecompilationOptions

from .disassembly_view import DisassemblyView
from .view import FunctionView

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace

log = logging.getLogger(__name__)


class CodeView(FunctionView):
    """
    A view to display pseudocode or source code. You should control this view by manipulating and observing its four
    ObjectContainers: .addr, .current_node, .codegen, and .function.
    """

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("pseudocode", workspace, default_docking_position, instance)

        self.base_caption = "Pseudocode"

        self._function: ObjectContainer | Function = ObjectContainer(None, "The function to decompile")
        self.current_node = ObjectContainer(None, "Current selected C-code node")
        self.addr: ObjectContainer | int = ObjectContainer(0, "Current cursor address")
        self.codegen: ObjectContainer | CStructuredCodeGenerator = ObjectContainer(
            None, "The currently-displayed codegen object"
        )

        self._last_function: Function | None = None
        self._textedit: QCCodeEdit | None = None
        self._doc: QCodeDocument | None = None
        self._options: QDecompilationOptions | None = None
        self.jump_history: JumpHistory = JumpHistory()
        self._nav_toolbar: NavToolbar | None = None
        self._view_selector: QComboBox | None = None

        self.vars_must_struct: set[str] = set()

        self._init_widgets()

        self._textedit.cursorPositionChanged.connect(self._on_cursor_position_changed)
        self._textedit.selectionChanged.connect(self._on_cursor_position_changed)
        self._textedit.mouse_double_clicked.connect(self._on_mouse_doubleclicked)
        self._function.am_subscribe(self._on_new_function)
        self.codegen.am_subscribe(self._on_codegen_changes)
        self.addr.am_subscribe(self._on_new_addr)
        self.current_node.am_subscribe(self._on_new_node)

    def _focus_core(self, focus: bool, focus_addr: int) -> None:
        if focus:
            self.focus()
        if focus_addr is not None:
            self.addr.am_obj = focus_addr
            self.addr.am_event(focus=True)

    #
    # Properties
    #

    @property
    def textedit(self):
        return self._textedit

    @property
    def document(self):
        return self._doc

    #
    # Public methods
    #

    def reload(self) -> None:
        if self.instance.project.am_none:
            return
        self._options.reload(force=True)
        self.vars_must_struct = set()

    def decompile(
        self,
        clear_prototype: bool = True,
        focus: bool = False,
        focus_addr=None,
        flavor: str = "pseudocode",
        reset_cache: bool = False,
        regen_clinic: bool = True,
    ) -> None:
        if self._function.am_none:
            return

        if clear_prototype:
            # clear the existing function prototype
            self._function.prototype = None
            self._function.ran_cca = False

        if reset_cache:
            self.instance.kb.structured_code.discard((self._function.addr, flavor))
            variables = self.instance.pseudocode_variable_kb.variables
            if variables.has_function_manager(self._function.addr):
                del variables[self._function.addr]

        def decomp_ready(*args, **kwargs) -> None:  # pylint:disable=unused-argument
            # this code is _partially_ duplicated from _on_new_function. be careful!
            available = self.instance.kb.structured_code.available_flavors(self._function.addr)
            self._update_available_views(available)
            if available:
                chosen_flavor = flavor if flavor in available else available[0]
                self.codegen.am_obj = self.instance.kb.structured_code[(self._function.addr, chosen_flavor)].codegen
                self.codegen.am_event(already_regenerated=True)
                self._focus_core(focus, focus_addr)
                if focus_addr is not None:
                    self.jump_history.record_address(focus_addr)
                else:
                    self.jump_history.record_address(self._function.am_obj.addr)

        def decomp(*_) -> None:
            job = DecompileFunctionJob(
                self.instance,
                self._function.am_obj,
                cfg=self.instance.cfg,
                options=self._options.option_and_values,
                optimization_passes=self._options.selected_passes,
                peephole_optimizations=self._options.selected_peephole_opts,
                vars_must_struct=self.vars_must_struct,
                on_finish=decomp_ready,
                blocking=True,
                regen_clinic=regen_clinic,
            )
            self.workspace.job_manager.add_job(job)

        if self._function.ran_cca is False:
            # run calling convention analysis for this function
            if self.instance._analysis_configuration:
                options = self.instance._analysis_configuration["varec"].to_dict()
            else:
                options = {}
            options["workers"] = 0
            varrec_job = VariableRecoveryJob(self.instance, **options, on_finish=decomp, func_addr=self._function.addr)
            self.workspace.job_manager.add_job(varrec_job)
        else:
            decomp()

    def highlight_chunks(self, chunks) -> None:
        extra_selections = []
        for start, end in chunks:
            sel = QTextEdit.ExtraSelection()
            sel.cursor = self._textedit.textCursor()
            sel.cursor.setPosition(start)
            sel.cursor.setPosition(end, QTextCursor.MoveMode.KeepAnchor)
            sel.format.setBackground(Conf.pseudocode_highlight_color)
            extra_selections.append(sel)
        self._textedit.setExtraSelections(extra_selections)

    #
    # Event callbacks
    #

    def _on_new_addr(
        self, already_moved: bool = False, focus: bool = False, **kwargs  # pylint:disable=unused-argument
    ) -> None:
        if already_moved:
            return

        # get closest node for ins
        new_text_pos = self._doc.find_closest_node_pos(self.addr.am_obj) if self._doc is not None else None

        if new_text_pos is not None:
            # set the new cursor position
            textedit = self.textedit
            cursor = textedit.textCursor()
            cursor.setPosition(new_text_pos)
            textedit.setTextCursor(cursor)
            if focus:
                textedit.setFocus()
                self._focus_core(True, None)
        else:
            # try to find the right function
            block_addr, _ = self.instance.cfb.floor_item(self.addr.am_obj)
            block = self.instance.cfg.get_any_node(block_addr)
            if block is not None:
                func = self.instance.kb.functions[block.function_address]
                if func is not self._function.am_obj:
                    self._function.am_obj = func
                    self._function.am_event(focus_addr=self.addr.am_obj, focus=focus)
                else:
                    log.error(
                        "There is a block which is in the current function but find_closest_node_pos failed on it"
                    )

    def _on_new_node(self, **kwargs) -> None:  # pylint: disable=unused-argument
        self.addr.am_obj = self._textedit.get_src_to_inst()
        self.addr.am_event(already_moved=True)

    # pylint: disable=unused-argument
    def _on_codegen_changes(self, already_regenerated: bool = False, event: str | None = None, **kwargs) -> None:
        """
        The callback function that triggers an update of the codegen.

        :param already_regenerated: True if we only want to re-render the text and do not intend to change the text.
                                    Setting it to True will ignore `event` and other parameters.
        :param event:               The event to perform. For example, "retype_variable" will cause a re-flow of
                                    variable types. Leaving it unspecified will lead to regeneration of text (which is
                                    the default behavior).
                                    Supported events: retype_variable
        :param kwargs:              Keyword arguments that are required in each event.
        :return:
        """
        if self.codegen.am_none:
            return

        old_lineno: int | None = None
        old_node: Any | None = None
        old_font = None
        if self._last_function is self._function.am_obj and self._doc is not None:
            # we are re-rendering the current function (e.g., triggered by a node renaming). the cursor should stay at
            # the same node
            old_cursor: QTextCursor = self._textedit.textCursor()
            old_pos = old_cursor.position()
            old_lineno = old_cursor.blockNumber()
            old_font = self._textedit.font()
            # TODO: If multiple instances of the node is referenced at the same line, we will always select the first
            #  node. We need to fix this by counting which node is selected.
            old_node = self._doc.get_node_at_position(old_pos)

        self._view_selector.setCurrentText(self.codegen.flavor)

        if already_regenerated:
            # do not regenerate text
            pass
        else:
            update_var_types = False
            if event == "retype_variable":
                dec = self.instance.project.analyses.Decompiler(
                    self._function.am_obj, variable_kb=self.instance.pseudocode_variable_kb, decompile=False
                )
                dec_cache = self.instance.kb.structured_code[(self._function.addr, "pseudocode")]
                new_codegen = dec.reflow_variable_types(
                    dec_cache.type_constraints,
                    dec_cache.func_typevar,
                    dec_cache.var_to_typevar or {},
                    dec_cache.codegen,
                )
                # update the cache
                dec_cache.codegen = new_codegen

                # update self
                self.codegen.am_obj = new_codegen
                update_var_types = True
            elif event == "retype_function":
                self.decompile(clear_prototype=False, reset_cache=True)
                update_var_types = True

            if not update_var_types:
                # regenerate text in the end
                self.codegen.regenerate_text()
            else:
                # trigger a full analysis
                self.codegen.cleanup()
                self.codegen._analyze()

        self._options.dirty = False
        if self._doc is None:
            self._doc = QCodeDocument(self.codegen)
            self._textedit.setDocument(self._doc)
        else:
            # only update the text. do not set a new document. this avoids scrolling the textedit up and down
            self._doc._codegen = self.codegen
            self._doc.setPlainText(self.codegen.text)

        if old_lineno is not None:
            the_block = self._doc.findBlockByNumber(old_lineno)
            new_cursor: QTextCursor = QTextCursor(the_block)
            self._textedit.setFont(old_font)
            if old_node is not None:
                # find the first node starting from the current position
                for pos in self.codegen.map_pos_to_node._posmap.irange(
                    minimum=new_cursor.position(), maximum=new_cursor.position() + the_block.length()
                ):
                    elem = self.codegen.map_pos_to_node._posmap[pos]
                    if elem.obj is old_node:
                        new_cursor.setPosition(pos)
            self._textedit.setTextCursor(new_cursor)

        if self.codegen.flavor == "pseudocode":
            self._options.show()
        else:
            self._options.hide()

    def _on_new_function(
        self, focus: bool = False, focus_addr=None, flavor=None, **kwargs
    ) -> None:  # pylint: disable=unused-argument
        # sets a new function. extra args are used in case this operation requires waiting for the decompiler
        if flavor is None:
            flavor = "pseudocode" if self.codegen.am_none else self.codegen.flavor

        if not self.codegen.am_none and self._last_function is self._function.am_obj:
            self._focus_core(focus, focus_addr)
            return
        available = self.instance.kb.structured_code.available_flavors(self._function.addr)
        self._update_available_views(available)
        should_decompile = True
        if available:
            chosen_flavor = flavor if flavor in available else available[0]
            cached = self.instance.kb.structured_code[(self._function.addr, chosen_flavor)].codegen
            if not isinstance(cached, DummyStructuredCodeGenerator):
                should_decompile = False
                self.codegen.am_obj = cached
                self.codegen.am_event()
                self._focus_core(focus, focus_addr)

        if should_decompile:
            self.decompile(focus=focus, focus_addr=focus_addr, flavor=flavor)
        self._last_function = self._function.am_obj

        console_view = self.workspace.view_manager.first_view_in_category("console")
        if console_view is not None:
            console_view.set_current_function(self._function.am_obj)

    def _on_cursor_position_changed(self) -> None:
        if self._doc is None:
            return

        cursor = self._textedit.textCursor()
        pos = cursor.position()
        selected_node = self._doc.get_node_at_position(pos)
        if selected_node is not None:
            # find all related text chunks and highlight them all
            chunks = self._doc.find_related_text_chunks(selected_node)
            # highlight these chunks
            self.highlight_chunks(chunks)
        else:
            self.highlight_chunks([])

        self.current_node.am_obj = selected_node
        self.current_node.am_event()

    def _navigate_to_function(self, from_addr: int, to_func: Function) -> None:
        self.jump_history.record_address(from_addr)
        self.jump_history.jump_to(to_func.addr)
        self.workspace.decompile_function(to_func, view=self)

    def _on_mouse_doubleclicked(self) -> None:
        if self._doc is None:
            return

        cursor = self._textedit.textCursor()
        pos = cursor.position()
        selected_node = self._doc.get_node_at_position(pos)
        if selected_node is not None:
            if isinstance(selected_node, CFunctionCall):
                # decompile this new function
                if selected_node.callee_func is not None:
                    self._navigate_to_function(selected_node.tags["ins_addr"], selected_node.callee_func)
            elif isinstance(selected_node, CConstant):
                # jump to highlighted constants
                if selected_node.reference_values is not None and selected_node.value is not None:
                    for v in selected_node.reference_values.values():
                        if isinstance(v, Function):
                            self._navigate_to_function(selected_node.tags["ins_addr"], v)
                            return
                    self.workspace.jump_to(selected_node.value)
            elif isinstance(selected_node, CVariable):
                var = selected_node.variable
                if var and isinstance(var, SimMemoryVariable):
                    self.workspace.jump_to(var.addr)

    def jump_to(self, addr: int, src_ins_addr=None) -> None:  # pylint:disable=unused-argument
        self.addr.am_obj = addr
        self.addr.am_event()

    def jump_back(self) -> None:
        addr = self.jump_history.backtrack()
        if addr is None:
            self.close()
        else:
            self.jump_to(addr)

    def popup_jumpto_dialog(self) -> None:
        view = self.workspace._get_or_create_view("disassembly", DisassemblyView)
        JumpTo(view, parent=self).exec_()
        view.decompile_current_function()

    def jump_forward(self) -> None:
        addr = self.jump_history.forwardstep()
        if addr is not None:
            self.jump_to(addr)

    def jump_to_history_position(self, pos: int) -> None:
        addr = self.jump_history.step_position(pos)
        if addr is not None:
            self.jump_to(addr)

    def keyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key_Tab:
            # Switch back to disassembly view
            self.workspace.jump_to(self.addr.am_obj)
            return True
        elif key == Qt.Key_G:
            # jump to window
            self.popup_jumpto_dialog()
            return True
        elif key == Qt.Key_Escape:
            self.jump_back()
            return True
        elif key == Qt.Key_Space and not self.codegen.am_none:
            flavor = self.codegen.flavor
            flavors = self.instance.kb.structured_code.available_flavors(self._function.addr)
            idx = flavors.index(flavor)
            newidx = (idx + 1) % len(flavors)
            self.codegen.am_obj = self.instance.kb.structured_code[(self._function.addr, flavors[newidx])].codegen
            self.codegen.am_event()
            return True

        return super().keyPressEvent(event)

    def _update_available_views(self, available) -> None:
        for _ in range(self._view_selector.count()):
            self._view_selector.removeItem(0)
        self._view_selector.addItems(available)
        self._view_selector.setVisible(len(available) >= 2)

    def _on_view_selector_changed(self, index) -> None:
        if not self._function.am_none:
            key = (self._function.addr, self._view_selector.itemText(index))
            self.codegen.am_obj = self.instance.kb.structured_code[key].codegen
            self.codegen.am_event()

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        window = QMainWindow()
        window.setWindowFlags(Qt.WindowType.Widget)

        # pseudo code text box
        self._textedit = QCCodeEdit(self)
        self._textedit.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByKeyboard | Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self._textedit.setLineWrapMode(QCCodeEdit.NoWrap)
        self._textedit.background = Conf.palette_base
        window.setCentralWidget(self._textedit)

        # decompilation options
        self._options = QDecompilationOptions(self, self.instance)
        options_dock = QDockWidget("Decompilation Options", self._options)
        window.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, options_dock)
        options_dock.setWidget(self._options)
        options_dock.setVisible(False)

        # status bar
        status_bar = QFrame()
        self._nav_toolbar = NavToolbar(
            self.jump_history, self.jump_back, self.jump_forward, self.jump_to_history_position, True, self
        )
        self._view_selector = QComboBox()
        self._view_selector.addItems(["Pseudocode"])
        self._view_selector.activated.connect(self._on_view_selector_changed)
        status_layout = QHBoxLayout()
        status_layout.addWidget(self._nav_toolbar.qtoolbar())
        status_layout.addStretch(0)
        options_toggle_btn = QToolButton()
        options_toggle_btn.setDefaultAction(options_dock.toggleViewAction())
        status_layout.addWidget(options_toggle_btn)
        status_layout.addWidget(self._view_selector)
        status_layout.setContentsMargins(3, 3, 3, 3)
        status_layout.setSpacing(3)
        status_bar.setLayout(status_layout)

        inner_layout = QHBoxLayout()
        inner_layout.addWidget(window)
        inner_layout.setContentsMargins(0, 0, 0, 0)
        inner_layout.setSpacing(0)
        inner_widget = QWidget()
        inner_widget.setLayout(inner_layout)

        outer_layout = QVBoxLayout()
        outer_layout.addWidget(status_bar)
        outer_layout.addWidget(inner_widget)
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.setSpacing(0)

        self.setLayout(outer_layout)

        self._textedit.focusWidget()

        self.workspace.plugins.instrument_code_view(self)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from .object_container import ObjectContainer

if TYPE_CHECKING:
    from collections.abc import Sequence


class BreakpointType(Enum):
    """
    Type of breakpoint.
    """

    Execute = 0
    Read = 1
    Write = 2


class Breakpoint:
    """
    A breakpoint / watchpoint.
    """

    __slots__ = ("type", "addr", "_size", "comment")

    def __init__(self, type_: BreakpointType, addr: int, size: int = 1, comment: str = "") -> None:
        self.type: BreakpointType = type_
        self.addr: int = addr
        self._size = size
        self.comment: str = comment

    @property
    def size(self):
        if self.type == BreakpointType.Execute:
            return 1
        return self._size

    @size.setter
    def size(self, v: int) -> None:
        self._size = v


class BreakpointManager:
    """
    Manager of breakpoints.
    """

    def __init__(self) -> None:
        self.breakpoints: ObjectContainer = ObjectContainer([], "List of breakpoints")

    def clear(self) -> None:
        self.breakpoints.clear()
        self.breakpoints.am_event()

    def add_breakpoint(self, bp: Breakpoint) -> None:
        self.breakpoints.append(bp)
        self.breakpoints.am_event(added=bp)

    def remove_breakpoint(self, bp: Breakpoint) -> None:
        self.breakpoints.remove(bp)
        self.breakpoints.am_event(removed=bp)

    def add_exec_breakpoint(self, addr: int) -> None:
        self.add_breakpoint(Breakpoint(BreakpointType.Execute, addr))

    def toggle_exec_breakpoint(self, addr: int) -> None:
        # is there a breakpoint at this address?
        found_bp = None
        for bp in self.breakpoints:
            if bp.type == BreakpointType.Execute and bp.addr == addr:
                # yes!
                found_bp = bp
                break

        if found_bp is None:
            self.add_exec_breakpoint(addr)
        else:
            self.remove_breakpoint(found_bp)

    def get_breakpoints_at(self, addr: int) -> Sequence[Breakpoint]:
        return [bp for bp in self.breakpoints if bp.addr <= addr < (bp.addr + bp.size)]
